import numpy as np
from numpy import sin, cos, pi, sqrt, tan, arccos, arcsin
try:
    from scipy.optimize import newton, minimize, curve_fit
    from scipy.signal import find_peaks, savgol_filter
except ImportError:
    _can_compute_eclipse_params = False
else:
    _can_compute_eclipse_params = True

import logging
logger = logging.getLogger("SOLVER")
logger.addHandler(logging.NullHandler())

# two-Gaussian model stuff

# HELPER FUNCTIONS

def ellipsoidal(phi, Aell, phi0):
    return 0.5*Aell*np.cos(4*np.pi*(phi-phi0))

def gaussian(phi, mu, d, sigma):
    return d*np.exp(-(phi-mu)**2/(2*sigma**2))

def gsum(phi, mu, d, sigma):
    gauss_sum = np.zeros(len(phi))
    for i in range(-2,3,1):
        gauss_sum += gaussian(phi,mu+i,d,sigma)
    return gauss_sum

# CHOICE OF MODELS

def const(phi, C):
    return C*np.ones(len(phi))

def ce(phi, C, Aell, phi0):
    return const(phi, C) - ellipsoidal(phi, Aell, phi0)

def cg(phi, C, mu, d,  sigma):
    return const(phi, C) - gsum(phi, mu, d, sigma)

def cge(phi, C, mu, d, sigma, Aell):
    return const(phi, C) - ellipsoidal(phi, Aell, mu) - gsum(phi, mu, d, sigma)

def cg12(phi, C, mu1, d1, sigma1, mu2, d2, sigma2):
    return const(phi, C) - gsum(phi, mu1, d1, sigma1) - gsum(phi, mu2, d2, sigma2)

def cg12e1(phi, C, mu1, d1, sigma1, mu2, d2, sigma2, Aell):
    return const(phi, C) - gsum(phi, mu1, d1, sigma1) - gsum(phi, mu2, d2, sigma2) - ellipsoidal(phi, Aell, mu1)

def cg12e2(phi, C, mu1, d1, sigma1, mu2, d2, sigma2, Aell):
    return const(phi, C) - gsum(phi, mu1, d1, sigma1) - gsum(phi, mu2, d2, sigma2) - ellipsoidal(phi, Aell, mu2)


# PREPROCESSING

def extend_phasefolded_lc(phases, fluxes, sigmas=None):

    #make new arrays that would span phase range -1 to 1:
    fluxes_extend = np.hstack((fluxes[(phases > 0)], fluxes, fluxes[phases < 0.]))
    phases_extend = np.hstack((phases[phases>0]-1, phases, phases[phases<0]+1))

    if sigmas is not None:
        sigmas_extend = np.hstack((sigmas[phases > 0], sigmas, sigmas[phases < 0.]))
    else:
        sigmas_extend = None

    return phases_extend, fluxes_extend, sigmas_extend


def find_eclipse(phases, fluxes):
    phase_min = phases[np.nanargmin(fluxes)]
    ph_cross = phases[fluxes - np.nanmedian(fluxes) > 0]
    # this part looks really complicated but it really only accounts for eclipses split
    # between the edges of the phase range - if a left/right edge is not found, we look for 
    # it in the phases on the other end of the range
    # we then mirror the value back on the side of the eclipse position for easier width computation
    try:
        arg_edge_left = np.argmin(np.abs(phase_min - ph_cross[ph_cross<phase_min]))
        edge_left = ph_cross[ph_cross<phase_min][arg_edge_left]
    except:
        arg_edge_left = np.argmin(np.abs((phase_min+1)-ph_cross[ph_cross<(phase_min+1)]))
        edge_left = ph_cross[ph_cross<(phase_min+1)][arg_edge_left]-1
    try:
        arg_edge_right = np.argmin(np.abs(phase_min-ph_cross[ph_cross>phase_min]))
        edge_right = ph_cross[ph_cross>phase_min][arg_edge_right]
    except:
        arg_edge_right = np.argmin(np.abs((phase_min-1)-ph_cross[ph_cross>(phase_min-1)]))
        edge_right = ph_cross[ph_cross>(phase_min-1)][arg_edge_right]+1
                            
    return phase_min, edge_left, edge_right


def estimate_eclipse_positions_widths(phases, fluxes, diagnose_init=False):
    pos1, edge1l, edge1r = find_eclipse(phases, fluxes)
    fluxes_sec = fluxes.copy()
    fluxes_sec[((phases > edge1l) & (phases < edge1r)) | ((phases > edge1l+1) | (phases < edge1r-1))] = np.nan
    pos2, edge2l, edge2r = find_eclipse(phases, fluxes_sec)
    

    if diagnose_init:
        import matplotlib.pyplot as plt
        plt.figure(figsize=(10,8))
        plt.plot(phases, fluxes, '.')
        plt.axhline(y=np.median(fluxes), c='orange')
        for i,x in enumerate([pos1, edge1l, edge1r]):
            ls = '-' if i==0 else '--'
            plt.axvline(x=x, c='r', ls=ls)
        for i,x in enumerate([pos2, edge2l, edge2r]):
            ls = '-' if i==0 else '--'
            plt.axvline(x=x, c='g', ls=ls)

    return {'ecl_positions': [pos1, pos2], 'ecl_widths': [edge1r-edge1l, edge2r-edge2l]}
                
# FITTING

def lnlike(y, yerr, ymodel):
    if yerr is not None:
        return -np.sum(np.log((2*np.pi)**0.5*yerr)+(y-ymodel)**2/(2*yerr**2))
    else:
        return -np.sum((y-ymodel)**2)


def bic(y, yerr, ymodel, nparams):
    if yerr is not None:
        return 2*lnlike(y,yerr,ymodel) - nparams*np.log(len(y))
    else:
        return lnlike(y, yerr, ymodel)


def fit_twoGaussian_models(phases, fluxes, sigmas=None):
    # setup the initial parameters

    # fit all of the models to the data
    twogfuncs = {'C': const, 'CE': ce, 'CG': cg, 'CGE': cge, 'CG12': cg12, 'CG12E1': cg12e1, 'CG12E2': cg12e2}

    C0 = fluxes.max()
    init_pos_w = estimate_eclipse_positions_widths(phases, fluxes)
    mu10, mu20 = init_pos_w['ecl_positions']
    sigma10, sigma20 = init_pos_w['ecl_widths']
    d10 = fluxes.max()-fluxes[np.argmin(np.abs(phases-mu10))]
    d20 = fluxes.max()-fluxes[np.argmin(np.abs(phases-mu20))]
    Aell0 = 0.001

    init_params = {'C': [C0,],
        'CE': [C0, Aell0, mu10],
        'CG': [C0, mu10, d10, sigma10],
        'CGE': [C0, mu10, d10, sigma10, Aell0],
        'CG12': [C0, mu10, d10, sigma10, mu20, d20, sigma20],
        'CG12E1': [C0, mu10, d10, sigma10, mu20, d20, sigma20, Aell0],
        'CG12E2': [C0, mu10, d10, sigma10, mu20, d20, sigma20, Aell0]}

    # parameters used frequently for bounds
    fmax = fluxes.max()
    fmin = fluxes.min()
    fdiff = fmax - fmin

    bounds = {'C': ((0),(fmax)),
        'CE': ((0, 1e-6, -0.5),(fmax, fdiff, 0.5)),
        'CG': ((0., -0.5, 0., 0.), (fmax, 0.5, fdiff, 0.5)),
        'CGE': ((0., -0.5, 0., 0., 1e-6),(fmax, 0.5, fdiff, 0.5, fdiff)),
        'CG12': ((0.,-0.5, 0., 0., -0.5, 0., 0.),(fmax, 0.5, fdiff, 0.5, 0.5, fdiff, 0.5)),
        'CG12E1': ((0.,-0.5, 0., 0., -0.5, 0., 0., 1e-6),(fmax, 0.5, fdiff, 0.5, 0.5, fdiff, 0.5, fdiff)),
        'CG12E2': ((0.,-0.5, 0., 0., -0.5, 0., 0., 1e-6),(fmax, 0.5, fdiff, 0.5, 0.5, fdiff, 0.5, fdiff))}

    fits = {}

    # extend light curve on phase range [-1,1]
    phases, fluxes, sigmas = extend_phasefolded_lc(phases, fluxes, sigmas)

    for key in twogfuncs.keys():
        try:
            fits[key] = curve_fit(twogfuncs[key], phases, fluxes, p0=init_params[key], sigma=sigmas, bounds=bounds[key])
        except Exception as err:
            logger.warning("2G model {} failed with error: {}".format(key, err))
            fits[key] = np.array([np.nan*np.ones(len(init_params[key]))])

    return fits


def compute_twoGaussian_models(fits, phases):
    twogfuncs = {'C': const, 'CE': ce, 'CG': cg, 'CGE': cge, 'CG12': cg12, 'CG12E1': cg12e1, 'CG12E2': cg12e2}

    models = {}

    for fkey in fits.keys():
        models[fkey] = twogfuncs[fkey](phases, *fits[fkey][0])

    return models


def compute_twoGaussian_models_BIC(models, phases, fluxes, sigmas):
    bics = {}
    nparams = {'C':1, 'CE':3, 'CG':4, 'CGE':5, 'CG12':7, 'CG12E1':8, 'CG12E2':8}

    for mkey in models.keys():
        bics[mkey] = bic(fluxes, sigmas, models[mkey], nparams[mkey])

    return bics


def fit_lc(phases, fluxes, sigmas):

    fits = fit_twoGaussian_models(phases, fluxes, sigmas)
    models = compute_twoGaussian_models(fits, phases)
    bics = compute_twoGaussian_models_BIC(models, phases, fluxes, sigmas)
    params = {'C': ['C'],
            'CE': ['C', 'Aell', 'mu1'],
            'CG': ['C', 'mu1', 'd1', 'sigma1'],
            'CGE': ['C', 'mu1', 'd1', 'sigma1', 'Aell'],
            'CG12': ['C', 'mu1', 'd1', 'sigma1', 'mu2', 'd2', 'sigma2'],
            'CG12E1': ['C', 'mu1', 'd1', 'sigma1', 'mu2', 'd2', 'sigma2', 'Aell'],
            'CG12E2': ['C', 'mu1', 'd1', 'sigma1', 'mu2', 'd2', 'sigma2', 'Aell']}

    best_fit = list(models.keys())[np.nanargmax(list(bics.values()))]
    return {'fits':fits, 'models':models, 'bics':bics, 'best_fit':best_fit, 'model_parameters': params}

# REFINING THE FIT

def two_line_model(values,breakp,x,y,sigma,edge='none'):
    k, n, a, b, c = values
    ymodel = np.zeros(len(x))
    if edge == 'left':
        ymodel[x<breakp] = k*x[x<breakp] + n
        ymodel[x>=breakp] = a*x[x>=breakp]**2+b*x[x>=breakp]+c
    elif edge == 'right':
        ymodel[x>breakp] = k*x[x>breakp] + n
        ymodel[x<=breakp] = a*x[x<=breakp]**2+b*x[x<=breakp]+c
    else:
        raise ValueError('Must provide value for edge orientation [\'left\', \'right\']')
    return np.sum((ymodel-y)**2/sigma**2)


def refine_eclipse_widths(phases, fluxes, sigmas, pos1, pos2, width1, width2):

    # to refine the region around the eclipses, we're taking half of the number of eclipse points
    # left and right from the current edge position
    mask1_left = (phases > pos1-width1) & (phases < pos1-0.05*width1)
    mask1_right = (phases > pos1+0.3*width1) & (phases < pos1+width1)
    mask2_left = (phases > pos2-width2) & (phases < pos2-0.05*width1)
    mask2_right = (phases > pos2+0.3*width1) & (phases < pos2+width2)

    eclipse_breaks = np.zeros(4)

    try:
        for i,mask in enumerate([mask1_left, mask1_right, mask2_left, mask2_right]):
            if i%2==0:
                edge='left'
            else:
                edge='right'
            breakpoints = np.linspace(phases[mask].min(), phases[mask].max(), len(phases[mask]))
            chis2 = np.zeros(len(breakpoints))
            for j,breakp in enumerate(breakpoints):
                sol = minimize(two_line_model, [0., 2., 1., 1., 2.], args=(breakp, phases[mask], fluxes[mask], sigmas[mask], edge))
                k, n, a, b, c = sol['x']
                x=phases[mask]
                ymodel = np.zeros(len(x))
                if edge == 'left':
                    ymodel[x<breakp] = k*x[x<breakp] + n
                    ymodel[x>=breakp] = a*x[x>=breakp]**2+b*x[x>=breakp]+c
                elif edge == 'right':
                    ymodel[x>breakp] = k*x[x>breakp] + n
                    ymodel[x<=breakp] = a*x[x<=breakp]**2+b*x[x<=breakp]+c
                else:
                    raise ValueError('Must provide value for edge orientation [\'left\', \'right\']')
                chis2[j] = np.sum((ymodel-fluxes[mask])**2/sigmas[mask]**2)
            eclipse_breaks[i] = breakpoints[np.argmin(chis2)]

        return eclipse_breaks
    except:
        logger.warning('Eclipse width refinement failed.')
        return [pos1-0.5*width1, pos1+0.5*width1, pos2-0.5*width2, pos2+0.5*width2]

# GEOMETRY SOLVER

def compute_eclipse_params(phases, fluxes, sigmas, fit_result=None, diagnose=False):

    if fit_result is None:
        fit_result = fit_lc(phases, fluxes, sigmas)
    best_fit = fit_result['best_fit']
    model_params = fit_result['model_parameters'][best_fit]

    sigma1 = fit_result['fits'][best_fit][0][model_params.index('sigma1')] if 'sigma1' in model_params else np.nan
    sigma2 = fit_result['fits'][best_fit][0][model_params.index('sigma2')] if 'sigma2' in model_params else np.nan
    mu1 = fit_result['fits'][best_fit][0][model_params.index('mu1')] if 'mu1' in model_params else np.nan
    mu2 = fit_result['fits'][best_fit][0][model_params.index('mu2')] if 'mu2' in model_params else np.nan
    C = fit_result['fits'][best_fit][0][model_params.index('C')]

    if not np.isnan(mu1) and not np.isnan(sigma1) and np.abs(sigma1) < 0.5:
        pos1 = mu1
        width1 = min(5.6*np.abs(sigma1), 0.5)
        depth1 = C - fluxes[np.argmin(np.abs(phases-pos1))]
    else:
        pos1 = np.nan
        width1 = np.nan
        depth1 = np.nan
    if not np.isnan(mu2) and not np.isnan(sigma2) and np.abs(sigma2) < 0.5:
        pos2 = mu2
        width2 = min(5.6*np.abs(sigma2), 0.5)
        depth2 = C - fluxes[np.argmin(np.abs(phases-pos2))]
    else:
        pos2 = np.nan
        width2 = np.nan
        depth2 = np.nan

    eclipse_edges = [pos1 - 0.5*width1, pos1+0.5*width1, pos2-0.5*width2, pos2+0.5*width2]
    # print('No eclipse refinement')
    # if not np.isnan(width1) and not np.isnan(width2) and not np.isnan(pos1) and not np.isnan(pos2):
    #     if np.abs(pos1-pos2) < width1 or np.abs(pos1-pos2) < width2:
    #         # in case of higly ellipsoidal systems, the eclipse positions aren't detected well
    #         # and need to be refined
    #         logger.warning('Poor two-Gaussian fit. Results potentially unreliable!')
    #         pos1 = phases_w[(phases_w > -0.25) & (phases_w < 0.25)][np.argmin(fluxes_w[(phases_w > -0.25) & (phases_w < 0.25)])]
    #         pos2 = phases_w[(phases_w > 0.25) & (phases_w < 0.75)][np.argmin(fluxes_w[(phases_w > 0.25) & (phases_w < 0.75)])]
    #         width1 = 0.5
    #         width2 = 0.5

    #     eclipse_edges = refine_eclipse_widths(phases_w, fluxes_w, sigmas_w, pos1, pos2, width1, width2)
    #     width1, width2 = eclipse_edges[1]-eclipse_edges[0], eclipse_edges[3]-eclipse_edges[2]
    # else:
    #     eclipse_edges = [np.nan, np.nan, np.nan, np.nan]


    if diagnose:
        phases_w, fluxes_w, sigmas_w = extend_phasefolded_lc(phases, fluxes, sigmas)
        twogfuncs = {'C': const, 'CE': ce, 'CG': cg, 'CGE': cge, 'CG12': cg12, 'CG12E1': cg12e1, 'CG12E2': cg12e2}
        [ecl1_l, ecl1_r, ecl2_l, ecl2_r] = eclipse_edges

        import matplotlib.pyplot as plt
        fig = plt.figure(figsize=(10,8))
        ax = fig.add_subplot(111)
        ax.plot(phases_w, fluxes_w, 'k.')
        plt.plot(phases_w, twogfuncs[best_fit](phases_w, *fit_result['fits'][best_fit][0]), '-', label=fit_result['best_fit'])
        lines = []
        lines.append(ax.axvline(x=pos1, c='#2B71B1', lw=2, label='primary'))
        lines.append(ax.axvline(x=pos2, c='#FF702F', lw=2, label='secondary'))
        lines.append(ax.axvline(x=ecl1_l, c='#2B71B1', lw=2, ls='--'))
        lines.append(ax.axvline(x=ecl1_r, c='#2B71B1', lw=2, ls='--'))
        lines.append(ax.axvline(x=ecl2_l, c='#FF702F', lw=2, ls='--'))
        lines.append(ax.axvline(x=ecl2_r, c='#FF702F', lw=2, ls='--'))
        drs = []
        for l,label in zip(lines,['pos1', 'pos2', 'ecl1_l', 'ecl1_r', 'ecl2_l', 'ecl2_r']):   
            dr = DraggableLine(l)
            dr.label = label
            dr.connect()   
            drs.append(dr) 
        ax.legend()
        plt.show(block=True)

        print('adjusting values')

        pos1 = drs[0].point.get_xdata()[0]
        pos2 = drs[1].point.get_xdata()[0]
        ecl1_l = drs[2].point.get_xdata()[0]
        ecl1_r = drs[3].point.get_xdata()[0]
        ecl2_l = drs[4].point.get_xdata()[0]
        ecl2_r = drs[5].point.get_xdata()[0]
        
        eclipse_edges = [ecl1_l, ecl1_r, ecl2_l, ecl2_r]


    return {
        'primary_width': width1,
        'secondary_width': width2,
        'primary_position': pos1,
        'secondary_position': pos2,
        'primary_depth': depth1,
        'secondary_depth': depth2,
        'eclipse_edges': eclipse_edges
    }

# t0 ESTIMATOR

def t0_from_geometry(phase_min, times, period=1, t0_supconj = 0, t0_near_times = True):

    delta_t0 = phase_min*period
    t0 = t0_supconj + delta_t0

    if t0_near_times:
        if t0 >= times.min() and t0 <= times.max():
            return t0
        else:
            return t0 + int((times.min()/period)+1)*(period)
    else:
        return t0



# ECCENTRICITY AND ARG OF PERIASTRON ESTIMATOR

def f (psi, sep): # used in pf_ecc_psi_w
    return psi - sin(psi) - 2*pi*sep

def df (psi, sep): # used in pf_ecc_psi_w
    return 1 - cos(psi) +1e-6

def ecc_w_from_geometry(sep, pwidth, swidth):

    if np.isnan(sep) or np.isnan(pwidth) or np.isnan(swidth):
        logger.warning('Cannot esimate eccentricty and argument of periastron: incomplete geometry information')
        return 0., pi/2
        
    # computation fails if sep<0, so we need to adjust for it here.
    if sep < 0:
        sep = 1+sep

    psi = newton(func=f, x0=(12*pi*sep)**(1./3), fprime=df, args=(sep,), maxiter=5000)
    # ecc = sqrt( (0.25*(tan(psi-pi))**2+(swidth-pwidth)**2/(swidth+pwidth)**2)/(1+0.25*(tan(psi-pi))**2) )
    ecc = (np.sin(0.5*(psi-pi))**2+((swidth-pwidth)/(swidth+pwidth))**2*np.cos(0.5*(psi-pi))**2)**0.5
    try:
        w1 = np.arcsin((swidth-pwidth)/(swidth+pwidth)/ecc)
        w2 = np.arccos((1-ecc**2)**0.5/ecc * np.tan(0.5*(psi-np.pi)))

        w = w2 if w1 >= 0 else 2*pi-w2
    except:
        w = pi/2
    return ecc, w


class DraggableLine:

    def __init__(self, p):
        self.point = p
        self.press = None

    def connect(self):
        self.cidpress = self.point.figure.canvas.mpl_connect('button_press_event', self.button_press_event)
        self.cidrelease = self.point.figure.canvas.mpl_connect('button_release_event', self.button_release_event)
        self.cidmotion = self.point.figure.canvas.mpl_connect('motion_notify_event', self.motion_notify_event)

    def disconnect(self):
        #disconnect all the stored connection ids
        self.point.figure.canvas.mpl_disconnect(self.cidpress)
        self.point.figure.canvas.mpl_disconnect(self.cidrelease)
        self.point.figure.canvas.mpl_disconnect(self.cidmotion)


    def button_press_event(self,event):
        if event.inaxes != self.point.axes:
            return
        contains = self.point.contains(event)[0]
        if not contains: return
        self.press = self.point.get_xdata(), event.xdata

    def button_release_event(self,event):
        self.press = None
        self.point.figure.canvas.draw()

    def motion_notify_event(self, event):
        if self.press is None: return
        if event.inaxes != self.point.axes: return
        xdata, xpress = self.press
        dx = event.xdata-xpress
        self.point.set_xdata(xdata+dx)
        self.point.figure.canvas.draw()



from scipy import stats as _stats
from scipy import interpolate as _interpolate
import numpy as _np

### custom subclasses of scipy.stats.rv_continuous

class delta_gen(_stats.rv_continuous):
    def _pdf(self, x):
        return _np.asarray(x==0.0, dtype='int')

    def _cdf(self, x):
        return _np.asarray(x>=0.0, dtype='int')

    def _ppf(self, q):
        return _np.zeros_like(q)

delta = delta_gen(name='delta')

class base_callable(object):
    def __init__(self, *args, **kwargs):
        return

    def __call__(self, x):
        raise NotImplementedError("{} must override __call__".format(self.__class__.__name__))

    def __mul__(self, other):
        # this is an UGLY hack to allow our callable to pass the scipy stats argument checks
        return self

    def __gt__(self, other):
        return True

class interpolate_callable(base_callable):
    def __init__(self, x, y):
        self._interpolator = _interpolate.interp1d(x, y, bounds_error=False, fill_value=0.0, assume_sorted=True)

    def __call__(self, x):
        return self._interpolator(x)

# class composite_callable(base_callable):
#     def __init__(self, d1_call, d2_call, math):
#         # math should be the math on the pdfs, not __and__/__or__
#         self._d1_call = d1_call
#         self._d2_call = d2_call
#         self._math = math
#
#    def __call__(self, x):
#        return getattr(self._d1_call(x), self._math)(self._d2_call(x))


# class generic_pdf_gen(_stats.rv_continuous):
#     """
#     this class takes simply a callable function to expose an arbitrary pdf,
#     and relies on the stats.rv_continuous infrastructure to compute and expose
#     everything else (cdf, ppf, etc).  This results in overheads which should
#     be avoided whenever possible, but does allow for full flexibility when it
#     isn't.
#
#     NOTE: it is important that the pdf_callable has an integral of 1
#     """
#     def _pdf(self, x, pdf_callable):
#         # TODO: I have ZERO idea why sometimes this is receiving an array with
#         # the interpolator and other times just gets the interpolator, but for
#         # now we'll just add another hack ;-)
#         if hasattr(pdf_callable, '__iter__'):
#             # if len(pdf_callable) > 1:
#             #     raise ValueError("could not parse arguments")
#             return pdf_callable[0](x)
#         return pdf_callable(x)

# class generic_cdf_gen(_stats.rv_continuous):
#     """
#     this class takes simply a callable function to expose an arbitrary cdf,
#     and relies on the stats.rv_continuous infrastructure to compute and expose
#     everything else (pdf, ppf, etc).  This results in overheads which should
#     be avoided whenever possible, but does allow for full flexibility when it
#     isn't.
#
#     NOTE: it is important that the cdf_callable starts at 0 (at -inf) and ends at 1 (at +inf)
#     """
#
#     def _cdf(self, x, cdf_callable):
#         # TODO: I have ZERO idea why sometimes this is receiving an array with
#         # the interpolator and other times just gets the interpolator, but for
#         # now we'll just add another hack ;-)
#         if hasattr(cdf_callable, '__iter__'):
#             # if len(cdf_callable) > 1:
#             #     raise ValueError("could not parse arguments")
#             return cdf_callable[0](x)
#         return cdf_callable(x)

class generic_pdf_cdf_ppf_gen(_stats.rv_continuous):
    """
    this class takes callable functions for each of pdf, cdf, and ppf.

    NOTE: it is important that the pdf_callable has an integral of 1
    NOTE: it is important that the cdf_callable starts at 0 (at -inf) and ends at 1 (at +inf)
    """
    def _pdf(self, x, pdf_callable, cdf_callable, ppf_callable):
        # TODO: I have ZERO idea why sometimes this is receiving an array with
        # the interpolator and other times just gets the interpolator, but for
        # now we'll just add another hack ;-)
        if hasattr(pdf_callable, '__iter__'):
            # if len(pdf_callable) > 1:
            #     raise ValueError("could not parse arguments")
            return pdf_callable[0](x)
        return pdf_callable(x)

    def _cdf(self, x, pdf_callable, cdf_callable, ppf_callable):
        # TODO: I have ZERO idea why sometimes this is receiving an array with
        # the interpolator and other times just gets the interpolator, but for
        # now we'll just add another hack ;-)
        if hasattr(cdf_callable, '__iter__'):
            # if len(cdf_callable) > 1:
            #     raise ValueError("could not parse arguments")
            return cdf_callable[0](x)
        return cdf_callable(x)

    def _ppf(self, x, pdf_callable, cdf_callable, ppf_callable):
        # TODO: I have ZERO idea why sometimes this is receiving an array with
        # the interpolator and other times just gets the interpolator, but for
        # now we'll just add another hack ;-)
        if hasattr(ppf_callable, '__iter__'):
            # if len(ppf_callable) > 1:
                # print(ppf_callable)
                # raise ValueError("could not parse arguments")
            return ppf_callable[0](x)
        return ppf_callable(x)

# generic_pdf = generic_pdf_gen(name='generic_pdf')
# generic_cdf = generic_cdf_gen(name='generic_cdf')
generic_pdf_cdf_ppf = generic_pdf_cdf_ppf_gen(name='generic_pdf_cdf_ppf')




from phoebe.parameters import *
from phoebe import conf

### NOTE: if creating new parameters, add to the _forbidden_labels list in parameters.py

def _comments_params(**kwargs):
    """
    """
    params = []

    params += [StringParameter(qualifier='comments', value=kwargs.get('comments', ''), description='User-provided comments for these solver-options.  Feel free to place any notes here - if not overridden, they will be copied to any resulting solutions.')]
    return params

def emcee(**kwargs):
    """
    Create a <phoebe.parameters.ParameterSet> for solver options for the
    emcee backend.  To use this backend, you must have emcee 3.0+ installed.

    To install emcee, see https://emcee.readthedocs.io

    If using this backend for solver, consider citing:
    * https://ui.adsabs.harvard.edu/abs/2013PASP..125..306F

    See also:
    * <phoebe.frontend.bundle.Bundle.references>
    * <phoebe.helpers.get_emcee_object_from_solution>
    * <phoebe.helpers.process_mcmc_chains_from_solution>

    Generally, this will be used as an input to the kind argument in
    <phoebe.frontend.bundle.Bundle.add_solver>.  If attaching through
    <phoebe.frontend.bundle.Bundle.add_solver>, all `**kwargs` will be
    passed on to set the values as described in the arguments below.  Alternatively,
    see <phoebe.parameters.ParameterSet.set_value> to set/change the values
    after creating the Parameters.

    For example:

    ```py
    b.add_solver('sampler.emcee')
    b.run_solver(kind='emcee')
    ```

    Parallelization support: emcee supports both MPI and multiprocessing.  If
    using MPI and nprocs > nwalkers and `compute` is a phoebe backend, then MPI
    will be handled at the compute-level (per-time).  In all other cases,
    parallelization is handled at the solver-level (per-model).

    The resulting solution (from <phoebe.frontend.bundle.Bundle.run_solver> or
    <phoebe.frontend.bundle.Bundle.export_solver> and <phoebe.frontend.bundle.Bundle.import_solution>)
    then exposes the raw-products from `emcee`, after which the following
    actions can be taken:

    * <phoebe.parameters.ParameterSet.plot> with `style` as one of:
        'corner', 'failed', 'lnprobabilities', 'trace'/'walks'.
    * <phoebe.frontend.bundle.Bundle.adopt_solution> to adopt the resulting
        posteriors in a distribution.  Use `adopt_values=True` (defaults to False)
        to adopt the face-values.  Use `trial_run=True` to see the adopted
        distributions and/or values without applying to the Bundle.
    * <phoebe.frontend.bundle.Bundle.get_distribution_collection> to access
        the multivariate distribution representation of the posteriors.
    * <phoebe.helpers.process_mcmc_chains> or <phoebe.helpers.process_mcmc_chains_from_solution>
        to manually access the `lnprobabilities` and `samples` after applying
        `burnin`, `thin`, `lnprob_cutoff`, and `adopt_parameters`/`adopt_inds`.

    Arguments
    ----------
    * `compute` (string, optional): compute options to use for forward model
    * `continue_from` (string, optional, default='None'): continue the MCMC run
        from an existing emcee solution.  Chains will be appended to existing
        chains (so it is safe to overwrite the existing solution).  If 'None',
        will start a new run using `init_from`.
    * `init_from` (list, optional, default=[]): only applicable if `continue_from`
        is 'None'.  distribution(s) to initialize samples from (all unconstrained
        parameters with attached distributions will be sampled/fitted, constrained
        parameters will be ignored, covariances will be respected)
    * `init_from_combine` (string, optional, default='first'): only applicable
        if `continue_from` is 'None' and `init_from` is not empty.  Method to use
        to combine multiple distributions from `init_from` for the same parameter.
        first: ignore duplicate entries and take the first in the init_from parameter.
        and: combine duplicate entries via AND logic, dropping covariances.
         or: combine duplicate entries via OR logic, dropping covariances.
    * `priors` (list, optional, default=[]): distribution(s) to use for priors
        (constrained and unconstrained parameters will be included, covariances
        will be respected except for distributions merge via `priors_combine`)
    * `priors_combine` (string, optional, default='and'): only applicable
        if `priors` is not empty.  Method to use to combine multiple distributions
        from `priors` for the same parameter.
        first: ignore duplicate entries and take the first in the priors parameter.
        and: combine duplicate entries via AND logic, dropping covariances.
        or: combine duplicate entries via OR logic, dropping covariances.
    * `nwalkers` (int, optional, default=16): only appicable if `continue_from`
        is 'None'.  Number of walkers.
    * `niters` (int, optional, default=100): Number of iterations.
    * `burnin_factor` (float, optional, default=2): factor of max(autocorr_times)
        to apply for burnin (burnin not applied until adopting the solution)
    * `thin_factor` (float, optional, default=0.5): factor of min(autocorr_times)
        to apply for thinning (thinning not applied until adopting the solution)
    * `progress_every_niters` (int, optional, default=0): Save the progress of
        the solution every n iterations.  The solution can only be recovered
        from an early termination by loading the bundle from a saved file and
        then calling <phoebe.frontend.bundle.Bundle.import_solution>(filename).
        The filename of the saved file will default to solution.ps.progress within
        <phoebe.frontend.bundle.Bundle.run_solver>, or the output filename provided
        to <phoebe.frontend.bundle.Bundle.export_solver> suffixed with .progress.
        If using detach=True within run_solver, attach job will load the progress
        and allow re-attaching until the job is completed.  If 0 will not save
        and will only return after completion.
    * `expose_failed` (bool, optional, default=True): only applicable if
        `continue_from` is 'None'. whether to expose dictionary of failed samples
        and their error messages.  Note: depending on the number of failed
        samples, this could add overhead.

    Returns
    --------
    * (<phoebe.parameters.ParameterSet>): ParameterSet of all newly created
        <phoebe.parameters.Parameter> objects.
    """
    params = _comments_params(**kwargs)

    params += [ChoiceParameter(qualifier='compute', value=kwargs.get('compute', 'None'), choices=['None'], description='compute options to use for forward model')]

    params += [ChoiceParameter(qualifier='continue_from', value=kwargs.get('continue_from', 'None'), choices=['None'], description='continue the MCMC run from an existing emcee solution.  Chains will be appended to existing chains (so it is safe to overwrite the existing solution).  If None, will start a new run using init_from.')]
    params += [SelectParameter(visible_if='continue_from:None', qualifier='init_from', value=kwargs.get('init_from', []), choices=[], description='distribution(s) to initialize samples from (all unconstrained parameters with attached distributions will be sampled/fitted, constrained parameters will be ignored, covariances will be respected)')]
    params += [ChoiceParameter(visible_if='continue_from:None,init_from:<notempty>', qualifier='init_from_combine', value=kwargs.get('init_from_combine', 'first'), choices=['first', 'and', 'or'], description='Method to use to combine multiple distributions from init_from for the same parameter.  first: ignore duplicate entries and take the first in the init_from parameter. and: combine duplicate entries via AND logic, dropping covariances.  or: combine duplicate entries via OR logic, dropping covariances.')]

    params += [SelectParameter(qualifier='priors', value=kwargs.get('priors', []), choices=[], description='distribution(s) to use for priors (constrained and unconstrained parameters will be included, covariances will be respected except for distributions merge via priors_combine)')]
    params += [ChoiceParameter(visible_if='priors:<notempty>', qualifier='priors_combine', value=kwargs.get('priors_combine', 'and'), choices=['first', 'and', 'or'], description='Method to use to combine multiple distributions from priors for the same parameter.  first: ignore duplicate entries and take the first in the priors parameter. and: combine duplicate entries via AND logic, dropping covariances.  or: combine duplicate entries via OR logic, dropping covariances.')]

    params += [IntParameter(visible_if='continue_from:None', qualifier='nwalkers', value=kwargs.get('nwalkers', 16), limits=(1,1e5), description='Number of walkers')]
    params += [IntParameter(qualifier='niters', value=kwargs.get('niters', 100), limits=(1,1e12), description='Number of iterations')]

    params += [FloatParameter(qualifier='burnin_factor', value=kwargs.get('burnin_factor', 2), default_unit=u.dimensionless_unscaled, limits=(1, 1000), description='factor of max(autocorr_times) to apply for burnin (burnin not applied until adopting the solution)')]
    params += [FloatParameter(qualifier='thin_factor', value=kwargs.get('thin_factor', 0.5), default_unit=u.dimensionless_unscaled, limits=(0.001, 1000), description='factor of min(autocorr_times) to apply for thinning (thinning not applied until adopting the solution)')]

    params += [IntParameter(qualifier='progress_every_niters', value=kwargs.get('progress_every_niters', 0), limits=(0,1e6), description='save the progress of the solution every n iterations.  The solution can only be recovered from an early termination by loading the bundle from a saved file and then calling b.import_solution(filename).  The filename of the saved file will default to solution.ps.progress within run_solver, or the output filename provided to export_solver suffixed with .progress.  If using detach=True within run_solver, attach job will load the progress and allow re-attaching until the job is completed.  If 0 will not save and will only return after completion.')]

    params += [BoolParameter(visible_if='continue_from:None', qualifier='expose_failed', value=kwargs.get('expose_failed', True), description='whether to expose dictionary of failed samples and their error messages.  Note: depending on the number of failed samples, this could add overhead.')]

    return ParameterSet(params)

def dynesty(**kwargs):
    """
    Create a <phoebe.parameters.ParameterSet> for solver options for the
    dynesty backend.  To use this backend, you must have dynesty installed.

    To install dynesty, see https://dynesty.readthedocs.io

    If using this backend for solver, consider citing:
    * https://ui.adsabs.harvard.edu/abs/2020MNRAS.493.3132S
    * https://ui.adsabs.harvard.edu/abs/2004AIPC..735..395S
    * https://projecteuclid.org/euclid.ba/1340370944

    and see:
    * https://dynesty.readthedocs.io/en/latest/#citations

    See also:
    * <phoebe.frontend.bundle.Bundle.references>
    * <phoebe.helpers.get_dynesty_object_from_solution>

    Generally, this will be used as an input to the kind argument in
    <phoebe.frontend.bundle.Bundle.add_solver>.  If attaching through
    <phoebe.frontend.bundle.Bundle.add_solver>, all `**kwargs` will be
    passed on to set the values as described in the arguments below.  Alternatively,
    see <phoebe.parameters.ParameterSet.set_value> to set/change the values
    after creating the Parameters.

    For example:

    ```py
    b.add_solver('sampler.dynesty')
    b.run_solver(kind='dynesty')
    ```

    Parallelization support: dynesty supports both MPI and multiprocessing, always
    at the solver-level (per-model).

    The resulting solution (from <phoebe.frontend.bundle.Bundle.run_solver> or
    <phoebe.frontend.bundle.Bundle.export_solver> and <phoebe.frontend.bundle.Bundle.import_solution>)
    then exposes the raw-products from `dynesty`, after which the following
    actions can be taken:

    * <phoebe.parameters.ParameterSet.plot> with `style` as one of:
        'corner', 'failed', 'trace', 'run'.
    * <phoebe.frontend.bundle.Bundle.adopt_solution> to adopt the resulting
        posteriors in a distribution.  Use `adopt_values=True` (defaults to False)
        to adopt the face-values.  Use `trial_run=True` to see the adopted
        distributions and/or values without applying to the Bundle.
    * <phoebe.frontend.bundle.Bundle.get_distribution_collection> to access
        the multivariate distribution representation of the posteriors.
    * <phoebe.helpers.get_dynesty_object> or <phoebe.helpers.get_dynesty_object_from_solution>
        to manually access the dynesty object which can then be passed to any
        [dynesty helper function](https://dynesty.readthedocs.io/en/latest/api.html#module-dynesty.results)
        which accepts the `results` object.


    Arguments
    ----------
    * `compute` (string, optional): compute options to use for forward model
    * `priors` (list, optional, default=[]): distribution(s) to use for priors
        (as dynesty samples directly from the prior, constrained parameters will
        be ignored, covariances will be dropped)
    * `priors_combine` (string, optional, default='and'): only applicable
        if `priors` is not empty.  Method to use to combine multiple distributions
        from `priors` for the same parameter.
        first: ignore duplicate entries and take the first in the priors parameter.
        and: combine duplicate entries via AND logic, dropping covariances.
        or: combine duplicate entries via OR logic, dropping covariances.
    * `nlive` (int, optional, default=100): number of live points.   Larger
        numbers result in a more finely sampled posterior (more accurate evidence),
        but also a larger number of iterations required to converge.
    * `maxiter` (int, optional, default=100): maximum number of iterations
    * `maxcall` (int, optional, default=1000): maximum number of calls (forward models)
    * `progress_every_niters` (int, optional, default=0): Save the progress of
        the solution every n iterations.  The solution can only be recovered
        from an early termination by loading the bundle from a saved file and
        then calling <phoebe.frontend.bundle.Bundle.import_solution>(filename).
        The filename of the saved file will default to solution.ps.progress within
        <phoebe.frontend.bundle.Bundle.run_solver>, or the output filename provided
        to <phoebe.frontend.bundle.Bundle.export_solver> suffixed with .progress.
        If using detach=True within run_solver, attach job will load the progress
        and allow re-attaching until the job is completed.  If 0 will not save
        and will only return after completion.
    * `expose_failed` (bool, optional, default=True): only applicable if
        `continue_from` is 'None'. whether to expose dictionary of failed samples
        and their error messages.  Note: depending on the number of failed
        samples, this could add overhead.


    Returns
    --------
    * (<phoebe.parameters.ParameterSet>): ParameterSet of all newly created
        <phoebe.parameters.Parameter> objects.
    """
    params = _comments_params(**kwargs)

    params += [ChoiceParameter(qualifier='compute', value=kwargs.get('compute', 'None'), choices=['None'], description='compute options to use for forward model')]

    params += [SelectParameter(qualifier='priors', value=kwargs.get('priors', []), choices=[], description='distribution(s) to use for priors (as dynesty samples directly from the prior, constrained parameters will be ignored, covariances will be dropped)')]
    params += [ChoiceParameter(visible_if='priors:<notempty>', qualifier='priors_combine', value=kwargs.get('priors_combine', 'and'), choices=['first', 'and', 'or'], description='Method to use to combine multiple distributions from priors for the same parameter. first: ignore duplicate entries and take the first in the priors parameter. and: combine duplicate entries via AND logic, dropping covariances.  or: combine duplicate entries via OR logic, dropping covariances.')]

    params += [IntParameter(qualifier='nlive', value=kwargs.get('nlive', 100), limits=(1,1e12), description='number of live points.   Larger numbers result in a more finely sampled posterior (more accurate evidence), but also a larger number of iterations required to converge.')]
    params += [IntParameter(qualifier='maxiter', value=kwargs.get('maxiter', 100), limits=(1,1e12), description='maximum number of iterations')]
    params += [IntParameter(qualifier='maxcall', value=kwargs.get('maxcall', 1000), limits=(1,1e12), description='maximum number of calls (forward models)')]

    params += [IntParameter(qualifier='progress_every_niters', value=kwargs.get('progress_every_niters', 0), limits=(0,1e6), description='save the progress of the solution every n iterations.  The solution can only be recovered from an early termination by loading the bundle from a saved file and then calling b.import_solution(filename).  The filename of the saved file will default to solution.ps.progress within run_solver, or the output filename provided to export_solver suffixed with .progress.  If using detach=True within run_solver, attach job will load the progress and allow re-attaching until the job is completed.  If 0 will not save and will only return after completion.')]

    params += [BoolParameter(qualifier='expose_failed', value=kwargs.get('expose_failed', True), description='whether to expose dictionary of failed samples and their error messages.  Note: depending on the number of failed samples, this could add overhead.')]


    return ParameterSet(params)




import numpy as np

from phoebe import c, u
import libphoebe


def rotfreq_to_omega(rotfreq, scale=c.R_sun.si.value, solar_units=False):
    """
    Translate from rotation frequency `rotfreq` to `omega`.

    NOTE: everything MUST be in consistent units according to `solar_units` bool

    Arguments
    ----------
    * `rotfreq`
    * `scale` (float, optional, default=c.R_sun.si.value)
    * `solar_units` (bool, optional, default=False): whether `scale` is provided
        in solar units or SI.

    Returns
    ---------
    * float
    """
    if solar_units:
        omega = rotfreq / (2*np.pi) / np.sqrt(c.GM_sun.to(u.solRad**3/u.d**2).value/scale**3)
    else:
        # then SI units
        omega = rotfreq / (2*np.pi) / np.sqrt(c.GM_sun.value/scale**3)

    # print "*** rotstar.rotfreq_to_omega", rotfreq, scale, solar_units, omega

    return omega

#
# def rpole2potential(rpole, rotfreq, solar_units=False):
#     """
#     Transforms polar radius to surface potential
#     """
#     if not solar_units:
#         rpole = rpole/c.R_sun.si.value
#     rpole_ = np.array([0., 0., rpole])
#     omega = rotfreq_to_omega(rotfreq, solar_units=solar_units)
#     pot =  libphoebe.rotstar_Omega(omega, rpole_)
#     # print "*** rotstar.rpole2potential", rpole, rotfreq, solar_units, omega, pot
#     return pot
#
#
# def potential2rpole(pot, rotfreq, solar_units=False):
#     """
#     Transforms surface potential to polar radius
#     """
#     omega = rotfreq_to_omega(rotfreq, scale=1.0, solar_units=solar_units)
#     # print "*** rotstar.potential2rpole", pot, rotfreq, solar_units, omega
#     rpole = libphoebe.rotstar_pole(omega, pot)
#     if solar_units:
#         return rpole
#     else:
#         return rpole*c.R_sun.si.value



from . import distl as _distl
from .distl import DistributionCollection, from_dict, from_json, from_file, get_random_seed, _has_astropy, _units, __version__, version # , sample_from_dists, sample_ppf_from_dists, logp_from_dists, sample_func_from_dists, plot_func_from_dists,
import numpy as _np
from .distl import BaseDistlObject, BaseDistribution, BaseAroundGenerator # for isinstance checking
import json as _json

try:
    import dill as _dill
except ImportError:
    _has_dill = False
else:
    _has_dill = True

name = 'distl'

def delta(value=0.0, unit=None, label=None, label_latex=None, wrap_at=None):
    """
    Create a <Delta> distribution.

    Arguments
    --------------
    * `value` (float or int, default=0.0): the value at which the delta function is True.
    * `unit` (astropy.units object, optional): the units of the provided values.
    * `label` (string, optional): a label for the distribution.  This is used
        for the x-label while plotting the distribution if `label_latex` is not provided,
        as well as a shorthand notation when creating a <Composite> distribution.
    * `label_latex` (string, optional): a latex label for the distribution.  This is used
        for the x-label while plotting.
    * `wrap_at` (float or False or None, optional, default=None): value to wrap all
        sampled values.  If None, will default to 0-2pi if `unit` is angular
        (0-360 for degrees), or 0-1 if `unit` is cycles.  If False, will not wrap.
        See <Delta.wrap_at> and <Delta.wrap> for more details.

    Returns
    --------
    * a <Delta> object
    """
    return _distl.Delta(value, unit=unit, label=label, label_latex=label_latex, wrap_at=wrap_at)


def uniform(low=0.0, high=1.0, unit=None, label=None, label_latex=None, wrap_at=None):
    """
    Create a <Uniform> distribution.

    Arguments
    --------------
    * `low` (float or int, default=0.0): the lower limit of the uniform distribution.
    * `high` (float or int, default=1.0): the upper limits of the uniform distribution.
        Must be higher than `low` unless `wrap_at` is provided or `unit`
        is provided as angular (rad, deg, cycles).
    * `unit` (astropy.units object, optional): the units of the provided values.
    * `label` (string, optional): a label for the distribution.  This is used
        for the x-label while plotting the distribution if `label_latex` is not provided,
        as well as a shorthand notation when creating a <Composite> distribution.
    * `label_latex` (string, optional): a latex label for the distribution.  This is used
        for the x-label while plotting.
    * `wrap_at` (float or False or None, optional, default=None): value to wrap all
        sampled values.  If None, will default to 0-2pi if `unit` is angular
        (0-360 for degrees), or 0-1 if `unit` is cycles.  If False, will not wrap.
        See <Uniform.wrap_at> and <Uniform.wrap> for more details.

    Returns
    --------
    * a <Uniform> object
    """
    return _distl.Uniform(low, high, unit=unit, label=label, label_latex=label_latex, wrap_at=wrap_at)

def boxcar(low=0.0, high=1.0, unit=None, label=None, label_latex=None, wrap_at=None):
    """
    Shortcut to <distl.uniform>.
    """
    return _distl.Uniform(low, high, unit=unit, label=label, label_latex=label_latex, wrap_at=wrap_at)


def gaussian(loc=0.0, scale=1.0, unit=None, label=None, label_latex=None, wrap_at=None):
    """
    Create a <Gaussian> distribution.

    Arguments
    --------------
    * `loc` (float or int, default=0.0): the central value of the gaussian distribution.
    * `scale` (float or int, default=1.0): the scale (sigma) of the gaussian distribution.
    * `unit` (astropy.units object, optional): the units of the provided values.
    * `label` (string, optional): a label for the distribution.  This is used
        for the x-label while plotting the distribution if `label_latex` is not provided,
        as well as a shorthand notation when creating a <Composite> distribution.
    * `label_latex` (string, optional): a latex label for the distribution.  This is used
        for the x-label while plotting.
    * `wrap_at` (float or False or None, optional, default=None): value to wrap all
        sampled values.  If None, will default to 0-2pi if `unit` is angular
        (0-360 for degrees), or 0-1 if `unit` is cycles.  If False, will not wrap.
        See <Gaussian.wrap_at> and <Gaussian.wrap> for more details.

    Returns
    --------
    * a <Gaussian> object
    """
    return _distl.Gaussian(loc, scale, unit=unit, label=label, label_latex=label_latex, wrap_at=wrap_at)

def normal(loc=0.0, scale=1.0, unit=None, label=None, label_latex=None, wrap_at=None):
    """
    Shortcut to <distl.gaussian>.
    """
    return _distl.Gaussian(loc, scale, unit=unit, label=label, label_latex=label_latex, wrap_at=wrap_at)

def histogram_from_bins(bins, density, unit=None, label=None, label_latex=None, wrap_at=None):
    """
    Create a <Histogram> distribution from binned data.

    See also:

    * <distl.histogram_from_data>

    Arguments
    --------------
    * `bins` (np.array object): the value of the bin-edges.  Must have one more
        entry than `density`.
    * `density` (np.array object): the value of the bin-densities.  Must have one
        less entry than `bins`.
    * `unit` (astropy.units object, optional): the units of the provided values.
    * `label` (string, optional): a label for the distribution.  This is used
        for the x-label while plotting the distribution if `label_latex` is not provided,
        as well as a shorthand notation when creating a <Composite> distribution.
    * `label_latex` (string, optional): a latex label for the distribution.  This is used
        for the x-label while plotting.
    * `wrap_at` (float or False or None, optional, default=None): value to wrap all
        sampled values.  If None, will default to 0-2pi if `unit` is angular
        (0-360 for degrees), or 0-1 if `unit` is cycles.  If False, will not wrap.
        See <Histogram.wrap_at> and <Histogram.wrap> for more details.

    Returns
    --------
    * a <Histogram> object
    """
    return _distl.Histogram(bins, density, unit=unit, label=label, label_latex=label_latex, wrap_at=wrap_at)

def histogram_from_data(data, bins=10, range=None, weights=None, unit=None, label=None, label_latex=None, wrap_at=None):
    """
    Create a <Histogram> distribution from data.

    See also:

    * <distl.histogram_from_bins>

    Arguments
    --------------
    * `data` (np.array object): 1D array of values.
    * `unit` (astropy.units object, optional): the units of the provided values.
    * `label` (string, optional): a label for the distribution.  This is used
        for the x-label while plotting the distribution if `label_latex` is not provided,
        as well as a shorthand notation when creating a <Composite> distribution.
    * `label_latex` (string, optional): a latex label for the distribution.  This is used
        for the x-label while plotting.
    * `wrap_at` (float or False or None, optional, default=None): value to wrap all
        sampled values.  If None, will default to 0-2pi if `unit` is angular
        (0-360 for degrees), or 0-1 if `unit` is cycles.  If False, will not wrap.
        See <Histogram.wrap_at> and <Histogram.wrap> for more details.

    Returns
    --------
    * a <Histogram> object
    """

    return _distl.Histogram.from_data(data, bins=bins, range=range,
                                        weights=weights, unit=unit,
                                        label=label, label_latex=label_latex,
                                        wrap_at=wrap_at)

def samples(samples, weights=None, bw_method=None, unit=None, label=None, label_latex=None, wrap_at=None):
    """
    Create a <Samples> distribution.

    Arguments
    --------------
    * `samples` (np.array object): an array of samples.  Note that any Nans
        will be removed.
    * `weights` (np.array object with length nsamples or None, optional, default=None):
        weights for each entry in `samples`.  NOTE: only supported with
        scipy 1.2+.
    * `bw_method` (string, float, or None, optional, default=None): passed
        directly to scipy.stats.gaussian_kde.  Only used for methods that
        rely on the KDE.
    * `unit` (astropy.units object, optional): the units of the provided values.
    * `label` (string, optional): a label for the distribution.  This is used
        for the x-label while plotting the distribution if `label_latex` is not provided,
        as well as a shorthand notation when creating a <Composite> distribution.
    * `label_latex` (string, optional): a latex label for the distribution.  This is used
        for the x-label while plotting.
    * `wrap_at` (float, None, or False, optional, default=None): value to
        use for wrapping.  If None and `unit` are angles, will default to
        2*pi (or 360 degrees).  If None and `unit` are cycles, will default
        to 1.0.

    Returns
    --------
    * a <Samples> object
    """
    return _distl.Samples(samples, weights, bw_method,
                         unit=unit, label=label, label_latex=label_latex,
                         wrap_at=wrap_at)


def function(func, args=[], kwargs={}, vectorized=True, hist_samples=None, unit=None, label=None, label_latex=None, wrap_at=None):
    """
    Create a <Function> distribution.


    Arguments
    ----------
    * `func`: callable function that accepts args and kwargs (as floats,
        after being sampled from any distribution objects)
    * `args` (list of distribution objects or floats): distribution objects
        or floats to pass as args to `func`.  Any items that are Distribution
        objects will be sampled and passed as floats.
    * `kwargs` (dictionary of distribution objects or floats): distribution
        objects or floats to pass as kwargs to `func`.  Any items that are
        Distribution objects will be sampled and passed as floats.
    * `vectorized` (bool, optional, default=True): whether `func` supports
        passing arrays to `args` and `kwargs`.
    * `hist_samples` (int, optional, default=None): number of samples to draw
        when generating the underlying <Histogram> distribution used for
        all probability calls.  If not provided or None, this will default
        to 1e6 if `vectorized` or 1e5 if not.  If `func` takes a long time
        or many samples in <Function.sample> are rejected and have to be
        re-drawn, it may be necessary to lower `hist_samples`.
    * `unit` (astropy.units object, optional): the units returned by `func`.
        Note that any Distribution objects in `args` and `kwargs` should be
        in the appropriate units (as the inputs and outputs to `func` are
        floats and not quantities)
    * `label` (string, optional): a label for the distribution.  This is used
        for the x-label while plotting the distribution if `label_latex` is not provided,
        as well as a shorthand notation when creating a <Composite> distribution.
    * `label_latex` (string, optional): a latex label for the distribution.  This is used
        for the x-label while plotting.
    * `wrap_at` (float, None, or False, optional, default=None): value to
        use for wrapping.  If None and `unit` are angles, will default to
        2*pi (or 360 degrees).  If None and `unit` are cycles, will default
        to 1.0.

    Returns
    ---------
    * a <Function> object
    """
    return _distl.Function(func, args, kwargs, vectorized, hist_samples,
                           unit=unit, label=label, label_latex=label_latex, wrap_at=wrap_at)


#### MULTIVARIATE DISTRIBUTJIONS ####

def mvgaussian(mean, cov, allow_singular=False,
               units=None, labels=None, labels_latex=None, wrap_ats=None):
    """
    Create a <MVGaussian> distribution.

    Arguments
    --------------
    * `mean` (float or int, default=0.0): the central value of the
        multivariate gaussian distribution.
    * `cov` (float or int, default=1.0): the covariance matrix of the multivariate
        gaussian distribution.
    * `allow_singular` (bool, optional, default=False): passed directly to
        scipy (see link above).
    * `units` (list of astropy.units objects, optional): the units of the provided values.
    * `labels` (list of strings, optional): labels for each dimension in the
        distribution.  This is used
        for the x-labels while plotting the distribution when `labels_latex`
        is not provided, as well as a shorthand
        notation when creating a <Composite> distribution.
    * `labels_latex` (list of strings, optional):  latex labels for each
        dimension in the distribution.  This is used for plotting the distribution.
    * `wrap_ats` (list of floats, None, or False, optional, default=None): values to
        use for wrapping.  If None and `unit` are angles, will default to
        2*pi (or 360 degrees).  If None and `unit` are cycles, will default
        to 1.0.

    Returns
    --------
    * a <MVGaussian> object
    """
    return _distl.MVGaussian(mean, cov, allow_singular=allow_singular,
                             units=units, labels=labels, labels_latex=labels_latex, wrap_ats=wrap_ats)

def mvhistogram_from_data(data, bins=10, range=None, weights=None,
                          units=None, labels=None, labels_latex=None, wrap_ats=None):
    """
    Create a <MVHistogram> object from data.

    Arguments
    ------------
    * `data` (array): input array of samples.  Passed to
        [np.histogramdd](https://numpy.org/doc/1.18/reference/generated/numpy.histogramdd.html)
    * `bins` (integer or array, optional, default=10): number of bins or
        bin edges.  Passed to [np.histogramdd](https://numpy.org/doc/1.18/reference/generated/numpy.histogramdd.html)
    * `weights` (array, optional, default=None): weights for each entry
        in `data`.  Passed to [np.histogramdd](https://numpy.org/doc/1.18/reference/generated/numpy.histogramdd.html)
    * `units` (list of astropy.units objects, optional): the units of the provided values.
    * `labels` (list of strings, optional): labels for each dimension in the
        distribution.  This is used
        for the x-labels while plotting the distribution when `labels_latex`
        is not provided, as well as a shorthand
        notation when creating a <Composite> distribution.
    * `labels_latex` (list of strings, optional):  latex labels for each
        dimension in the distribution.  This is used for plotting the distribution.
    * `wrap_ats` (list of floats, None, or False, optional, default=None): values to
        use for wrapping.  If None and `unit` are angles, will default to
        2*pi (or 360 degrees).  If None and `unit` are cycles, will default
        to 1.0.

    Returns
    ----------
    * a <MVHistogram> object
    """
    return _distl.MVHistogram.from_data(data, bins=bins, range=range, weights=weights,
                                        units=units, labels=labels, labels_latex=labels_latex, wrap_ats=wrap_ats)


def mvsamples(samples, weights=None, bw_method=None,
              units=None, labels=None, labels_latex=None, wrap_ats=None):
    """
    Create a <MVSamples> distribution.

    Arguments
    --------------
    * `samples` (np.array object with shape (nsamples, <MVSamples.ndimensions>)):
        the samples.
    * `weights` (np.array object with shape (nsamples) or None, optional, default=None):
        weights for each entry in `samples`.  NOTE: only supported with scipy
        version 1.2+.
    * `bw_method` (string, float, or None, optional, default=None): passed
        directly to scipy.stats.gaussian_kde.  Only used for methods that
        rely on the KDE.
    * `units` (list of astropy.units objects, optional): the units of the provided values.
    * `labels` (list of strings, optional): labels for each dimension in the
        distribution.  This is used
        for the x-labels while plotting the distribution when `labels_latex`
        is not provided, as well as a shorthand
        notation when creating a <Composite> distribution.
    * `labels_latex` (list of strings, optional):  latex labels for each
        dimension in the distribution.  This is used for plotting the distribution.
    * `wrap_ats` (list of floats, None, or False, optional, default=None): values to
        use for wrapping.  If None and `unit` are angles, will default to
        2*pi (or 360 degrees).  If None and `unit` are cycles, will default
        to 1.0.

    Returns
    --------
    * an <MVSamples> object
    """
    return _distl.MVSamples(samples, weights, bw_method, units=units, labels=labels, labels_latex=labels_latex, wrap_ats=wrap_ats)

#### GENERATORS ####
def gaussian_around(scale, value=None,
                    unit=None, label=None, label_latex=None, wrap_at=None):
    """
    Create a <Gaussian_Around> object which, when called, will resolve
    to a <Gaussian> object around a given central value.

    Arguments
    --------------
    * `scale` (float or int, default=1.0): the scale (sigma) of the gaussian
        distribution.
    * `value` (float, optional, default=None): the current face-value.
    * `unit` (astropy.units object, optional): the units of the provided values.
    * `label` (string, optional): a label for the distribution.  This is used
        for the x-label while plotting the distribution if `label_latex` is not provided,
        as well as a shorthand notation when creating a <Composite> distribution.
    * `label_latex` (string, optional): a latex label for the distribution.  This is used
        for the x-label while plotting.
    * `wrap_at` (float, None, or False, optional, default=None): value to
        use for wrapping.  If None and `unit` are angles, will default to
        2*pi (or 360 degrees).  If None and `unit` are cycles, will default
        to 1.0.
    Returns
    --------
    * a <Gaussian_Around> object
    """

    return _distl.Gaussian_Around(scale, value, unit, label, label_latex, wrap_at)

def uniform_around(width, value=None,
                   unit=None, label=None, label_latex=None, wrap_at=None):
    """
    Create a <Uniform_Around> object which, when called, will resolve
    to a <Uniform> object around a given central value.

    Arguments
    --------------
    * `width` (float): the width of the resulting <Uniform> object (<Uniform.low>
        and <Uniform.high> will be set based on the current value and `width`).
    * `value` (float, optional, default=None): the current face-value.
    * `unit` (astropy.units object, optional): the units of the provided values.
    * `label` (string, optional): a label for the distribution.  This is used
        for the x-label while plotting the distribution if `label_latex` is not provided,
        as well as a shorthand notation when creating a <Composite> distribution.
    * `label_latex` (string, optional): a latex label for the distribution.  This is used
        for the x-label while plotting.
    * `wrap_at` (float, None, or False, optional, default=None): value to
        use for wrapping.  If None and `unit` are angles, will default to
        2*pi (or 360 degrees).  If None and `unit` are cycles, will default
        to 1.0.

    Returns
    -----------
    * a <Uniform_Around> object.
    """

    return _distl.Uniform_Around(width, value, unit, label, label_latex, wrap_at)

def delta_around(value=None, unit=None, label=None, label_latex=None, wrap_at=None):
    """
    Create a <Delta_Around> object which, when called, will resolve
    to a <Delta> object around a given central value.

    Arguments
    --------------
    * `value` (float, optional, default=None): the current face-value.
    * `unit` (astropy.units object, optional): the units of the provided values.
    * `label` (string, optional): a label for the distribution.  This is used
        for the x-label while plotting the distribution if `label_latex` is not provided,
        as well as a shorthand notation when creating a <Composite> distribution.
    * `label_latex` (string, optional): a latex label for the distribution.  This is used
        for the x-label while plotting.
    * `wrap_at` (float, None, or False, optional, default=None): value to
        use for wrapping.  If None and `unit` are angles, will default to
        2*pi (or 360 degrees).  If None and `unit` are cycles, will default
        to 1.0.

    Returns
    --------
    * a <Delta> object
    """
    return _distl.Delta_Around(unit, value, label, label_latex, wrap_at)




from phoebe.parameters import *
from phoebe import conf
from phoebe import geomspace as _geomspace
from phoebe import linspace as _linspace

from distutils.version import StrictVersion
import numpy as np


### NOTE: if creating new parameters, add to the _forbidden_labels list in parameters.py

def _comments_params(**kwargs):
    """
    """
    params = []

    params += [StringParameter(qualifier='comments', value=kwargs.get('comments', ''), description='User-provided comments for these solver-options.  Feel free to place any notes here - if not overridden, they will be copied to any resulting solutions.')]
    return params


def lc_periodogram(**kwargs):
    """
    Create a <phoebe.parameters.ParameterSet> for solver options for the
    light curve periodogram period estimator using astropy.  See
    https://docs.astropy.org/en/stable/timeseries/bls.html or
    https://docs.astropy.org/en/stable/timeseries/lombscargle.html for more details.

    NOTE: this requires astropy 3.2+, which in turn requires python 3.  If these
    requirements are not met, an error will be raised when attempting to call
    <phoebe.frontend.bundle.Bundle.run_solver>.

    The input light curve datasets (`lc_datasets`) are each normalized
    according to `lc_combine` and then combined.
    These combined data are then sent to the respective periodgram
    `algorithm` and the resulting period
    corresponding to the strongest peak is proposed as an adopted value.  In
    addition, the periodgram itself is exposed in the solution and available
    for plotting via <phoebe.parameters.ParameterSet.plot>.

    Generally, this will be used as an input to the kind argument in
    <phoebe.frontend.bundle.Bundle.add_solver>.  If attaching through
    <phoebe.frontend.bundle.Bundle.add_solver>, all `**kwargs` will be
    passed on to set the values as described in the arguments below.  Alternatively,
    see <phoebe.parameters.ParameterSet.set_value> to set/change the values
    after creating the Parameters.

    For example:

    ```py
    b.add_solver('estimator.lc_periodogram')
    b.run_solver(kind='lc_periodogram')
    ```

    Arguments
    ----------
    * `algorithm` (string, optional, default='bls'): algorithm to use to create
        the periodogram.  bls: BoxLeastSquares, ls: LombScargle.
    * `lc_datasets` (string or list, optional, default='*'): Light curve
        dataset(s) to use to run the periodogram algorithm.
    * `lc_combine` (string, optional, default='median'): How to normalize each
        light curve prior to combining.
    * `component` (string, optional, default=top-level orbit): Component to
        apply the found period.
    * `sample_mode` (string, optional, default='auto'): Whether to automatically
        determine sampling periods/frequencies ('auto') or set manually ('manual').
    * `sample_periods` (array, optional, default=[]): only applicable if
        `sample_mode` is 'manual'.  Manual period grid for sampling the periodogram.
        Note: if `algorithm` is 'ls', these will be converted to frequencies and
        will be more efficient if sampled evenly in frequency space (consider
        using <phoebe.invspace> instead of <phoebe.linspace>).
    * `duration` (array, optional, default=geomspace(0.01,0.3,5)): only applicable
        if `algorithm` is 'bls'.  The set of durations (in phase-space) that
        will be considered.  See
        https://docs.astropy.org/en/stable/api/astropy.timeseries.BoxLeastSquares.html
    * `objective` (string, optional, default='likelihood'): only applicable if
        `algorithm` is 'bls'.  Objective to use for the periodogram.  See
        https://docs.astropy.org/en/stable/timeseries/bls.html#objectives
    * `minimum_n_cycles` (int, optional, default=3): only applicable if
        `algorithm` is 'bls' and `sample_mode` is 'auto'.  Minimum number of
        cycles/eclipses.  This is passed directly to autopower as
        'minimum_n_transit'. See
        https://docs.astropy.org/en/stable/api/astropy.timeseries.BoxLeastSquares.html#astropy.timeseries.BoxLeastSquares.autopower
    * `samples_per_peak` (int, optional, default=10): only applicable if
        `algorithm` is 'ls' and `sample_mode` is 'auto'.  The approximate number
        of desired samples across the typical peak.  This is passed directly to
        autopower. See
        https://docs.astropy.org/en/stable/api/astropy.timeseries.LombScargle.html#astropy.timeseries.LombScargle.autopower
    * `nyquist_factor` (int, optional, default=5): only applicable if
        `algorithm` is 'ls' and `sample_mode` is 'auto'.  The multiple of the
        average nyquist frequency used to choose the maximum frequency.  This is
        passed directly to autopower. See
        https://docs.astropy.org/en/stable/api/astropy.timeseries.LombScargle.html#astropy.timeseries.LombScargle.autopower


    Returns
    --------
    * (<phoebe.parameters.ParameterSet>): ParameterSet of all newly created
        <phoebe.parameters.Parameter> objects.
    """
    params = _comments_params(**kwargs)

    params += [ChoiceParameter(qualifier='algorithm', value=kwargs.get('algorithm', 'bls'), choices=['bls', 'ls'], description='Algorithm to use to create the periodogram.  bls: BoxLeastSquares, ls: LombScargle.')]

    params += [SelectParameter(qualifier='lc_datasets', value=kwargs.get('lc_datasets', '*'), choices=[], description='Light curve dataset(s) to use to run the periodogram algorithm')]
    params += [ChoiceParameter(visible_if='lc_datasets:<plural>', qualifier='lc_combine', value=kwargs.get('lc_combine', 'median'), choices=['median', 'max'], advanced=True, description='How to normalize each light curve prior to combining.')]

    params += [ChoiceParameter(qualifier='component', value=kwargs.get('component', ''), choices=[''], description='Component to apply the found period')]

    params += [ChoiceParameter(qualifier='sample_mode', value=kwargs.get('sample_mode', 'auto'), choices=['auto', 'manual'], description='Whether to automatically determine sampling periods/frequencies or set manually')]

    params += [FloatArrayParameter(visible_if='sample_mode:manual', qualifier='sample_periods', value=kwargs.get('sample_periods', []), default_unit=u.d, description='Manual period grid for sampling the periodogram.  Note: if algorithm is \'ls\', these will be converted to frequencies and will be more efficient if sampled evenly in frequency space (consider using phoebe.invspace instead of phoebe.linspace).')]

    # BLS-specific options
    params += [FloatArrayParameter(visible_if='algorithm:bls', qualifier='duration', value=kwargs.get('duration', [0.1]), default_unit=u.d, description='The set of durations (in time-space) that will be considered.  See https://docs.astropy.org/en/stable/api/astropy.timeseries.BoxLeastSquares.html')]
    params += [ChoiceParameter(visible_if='algorithm:bls', qualifier='objective', value=kwargs.get('objective', 'likelihood'), choices=['likelihood', 'snr'], description='Objective to use for the periodogram.  See https://docs.astropy.org/en/stable/timeseries/bls.html#objectives')]
    params += [IntParameter(visible_if='sample_mode:auto,algorithm:bls', qualifier='minimum_n_cycles', value=kwargs.get('minimum_n_cycles', 3), advanced=True, limits=(1,None), description='Minimum number of cycles/eclipses.  This is passed directly to autopower as \'minimum_n_transit\'. See https://docs.astropy.org/en/stable/api/astropy.timeseries.BoxLeastSquares.html#astropy.timeseries.BoxLeastSquares.autopower')]


    # LS-specific options
    params += [IntParameter(visible_if='sample_mode:auto,algorithm:ls', qualifier='samples_per_peak', value=kwargs.get('samples_per_peak', 10), advanced=True, limits=(1,None), description='The approximate number of desired samples across the typical peak.  This is passed directly to autopower. See https://docs.astropy.org/en/stable/api/astropy.timeseries.LombScargle.html#astropy.timeseries.LombScargle.autopower')]
    params += [IntParameter(visible_if='sample_mode:auto,algorithm:ls', qualifier='nyquist_factor', value=kwargs.get('nyquist_factor', 5), advanced=True, limits=(1,None), description='The multiple of the average nyquist frequency used to choose the maximum frequency.  This is passed directly to autopower. See https://docs.astropy.org/en/stable/api/astropy.timeseries.LombScargle.html#astropy.timeseries.LombScargle.autopower')]

    return ParameterSet(params)

def rv_periodogram(**kwargs):
    """
    Create a <phoebe.parameters.ParameterSet> for solver options for the
    radial velocity periodogram period estimator using astropy.  See
    https://docs.astropy.org/en/stable/timeseries/lombscargle.html for more details.

    NOTE: this requires astropy 3.2+, which in turn requires python 3.  If these
    requirements are not met, an error will be raised when attempting to call
    <phoebe.frontend.bundle.Bundle.run_solver>.

    The input radial velocity datasets (`rv_datasets`) are combined and then
    normalized by the absolute maximum value for the primary and secondary star
    independently, with the secondary then mirrored.  These combined data are
    then sent to the respective periodgram `algorithm` and the resulting period
    corresponding to the strongest peak is proposed as an adopted value.  In
    addition, the periodgram itself is exposed in the solution and available
    for plotting via <phoebe.parameters.ParameterSet.plot>.

    Generally, this will be used as an input to the kind argument in
    <phoebe.frontend.bundle.Bundle.add_solver>.  If attaching through
    <phoebe.frontend.bundle.Bundle.add_solver>, all `**kwargs` will be
    passed on to set the values as described in the arguments below.  Alternatively,
    see <phoebe.parameters.ParameterSet.set_value> to set/change the values
    after creating the Parameters.

    For example:

    ```py
    b.add_solver('estimator.rv_periodogram')
    b.run_solver(kind='rv_periodogram')
    ```

    Arguments
    ----------
    * `algorithm` (string, optional, default='bls'): algorithm to use to create
        the periodogram.  ls: LombScargle.
    * `rv_datasets` (string or list, optional, default='*'): Radial velocity
        dataset(s) to use to run the periodogram algorithm.
    * `component` (string, optional, default=top-level orbit): Component to
        apply the found period.
    * `sample_mode` (string, optional, default='auto'): Whether to automatically
        determine sampling periods/frequencies ('auto') or set manually ('manual').
    * `sample_periods` (array, optional, default=[]): only applicable if
        `sample_mode` is 'manual'.  Manual period grid for sampling the periodogram.
        Note: if `algorithm` is 'ls', these will be converted to frequencies and
        will be more efficient if sampled evenly in frequency space (consider
        using <phoebe.invspace> instead of <phoebe.linspace>).
    * `samples_per_peak` (int, optional, default=10): only applicable if
        `algorithm` is 'ls' and `sample_mode` is 'auto'.  The approximate number
        of desired samples across the typical peak.  This is passed directly to
        autopower. See
        https://docs.astropy.org/en/stable/api/astropy.timeseries.LombScargle.html#astropy.timeseries.LombScargle.autopower
    * `nyquist_factor` (int, optional, default=5): only applicable if
        `algorithm` is 'ls' and `sample_mode` is 'auto'.  The multiple of the
        average nyquist frequency used to choose the maximum frequency.  This is
        passed directly to autopower. See
        https://docs.astropy.org/en/stable/api/astropy.timeseries.LombScargle.html#astropy.timeseries.LombScargle.autopower



    Returns
    --------
    * (<phoebe.parameters.ParameterSet>): ParameterSet of all newly created
        <phoebe.parameters.Parameter> objects.
    """
    params = _comments_params(**kwargs)

    params += [ChoiceParameter(qualifier='algorithm', value=kwargs.get('algorithm', 'ls'), choices=['ls'], description='Algorithm to use to create the periodogram.  ls: LombScargle.')]

    params += [SelectParameter(qualifier='rv_datasets', value=kwargs.get('rv_datasets', '*'), choices=[], description='Radial velocity dataset(s) to use to run the periodgram algorithm')]

    params += [ChoiceParameter(qualifier='component', value=kwargs.get('component', ''), choices=[''], description='Component to apply the found period')]

    params += [ChoiceParameter(qualifier='sample_mode', value=kwargs.get('sample_mode', 'auto'), choices=['auto', 'manual'], description='Whether to automatically determine sampling periods/frequencies or set manually')]

    params += [FloatArrayParameter(visible_if='sample_mode:manual', qualifier='sample_periods', value=kwargs.get('sample_periods', []), default_unit=u.d, description='Manual period grid for sampling the periodogram.  Note: if algorithm is \'ls\', these will be converted to frequencies and will be more efficient if sampled evenly in frequency space (consider using phoebe.invspace instead of phoebe.linspace).')]

    # LS-specific options
    params += [IntParameter(visible_if='sample_mode:auto,algorithm:ls', qualifier='samples_per_peak', value=kwargs.get('samples_per_peak', 10), advanced=True, limits=(1,None), description='The approximate number of desired samples across the typical peak.  This is passed directly to autopower. See https://docs.astropy.org/en/stable/api/astropy.timeseries.LombScargle.html#astropy.timeseries.LombScargle.autopower')]
    params += [IntParameter(visible_if='sample_mode:auto,algorithm:ls', qualifier='nyquist_factor', value=kwargs.get('nyquist_factor', 5), advanced=True, limits=(1,None), description='The multiple of the average nyquist frequency used to choose the maximum frequency.  This is passed directly to autopower. See https://docs.astropy.org/en/stable/api/astropy.timeseries.LombScargle.html#astropy.timeseries.LombScargle.autopower')]

    return ParameterSet(params)


def lc_geometry(**kwargs):
    """
    Create a <phoebe.parameters.ParameterSet> for solver options for the
    light curve geometry esimator.

    The input light curve datasets (`lc_datasets`) are each normalized
    according to `lc_combine` and then combined.
    These combined data are then fitted with a 2-gaussian model
    which is used to help determine phases of eclipse minima, ingress, and
    egress.  These are then used to estimate and propose values for `ecc`, `per0`,
    `t0_supconj` for the corresponding `orbit` as well as `mask_phases` (not included in `adopt_parameters`
    by default).  If `expose_model` is True, the 2-gaussian model and the phases of minima,
    ingress, and egress are exposed in the solution and available for
    plotting with <phoebe.parameters.ParameterSet.plot>.

    Generally, this will be used as an input to the kind argument in
    <phoebe.frontend.bundle.Bundle.add_solver>.  If attaching through
    <phoebe.frontend.bundle.Bundle.add_solver>, all `**kwargs` will be
    passed on to set the values as described in the arguments below.  Alternatively,
    see <phoebe.parameters.ParameterSet.set_value> to set/change the values
    after creating the Parameters.

    For example:

    ```py
    b.add_solver('estimator.lc_geometry')
    b.run_solver(kind='lc_geometry')
    ```

    Arguments
    ----------
    * `lc_datasets` (string or list, optional, default='*'): Light curve
        dataset(s) to use to extract eclipse geometry
    * `lc_combine` (string, optional, default='median'): How to normalize each
        light curve prior to combining.
    * `phase_bin` (bool, optional, default=True): Bin the input observations (
        see `phase_nbins`) if more than 2*phase_nbins.  NOTE: input observational
        sigmas will be ignored during binning.
    * `phase_nbins` (int, optional, default=500): Number of bins to use during
        phase binning input observations
        (will only be applied if len(times) > 2*`phase_nbins`).  Only applicable
        if `phase_bin` is True.
    * `orbit` (string, optional, default=top-level orbit): Orbit to use for
        phasing the light curve referenced in the `lc_datasets` parameter
    * `t0_near_times` (bool, optional, default=True): Whether the returned value
        for t0_supconj should be forced to be in the range of the referenced
        observations.
    * `expose_model` (bool, optional, default=True): Whether to expose the
        2-gaussian analytical models in the solution

    Returns
    --------
    * (<phoebe.parameters.ParameterSet>): ParameterSet of all newly created
        <phoebe.parameters.Parameter> objects.
    """
    params = _comments_params(**kwargs)

    params += [SelectParameter(qualifier='lc_datasets', value=kwargs.get('lc_datasets', '*'), choices=[], description='Light curve dataset(s) to use to extract eclipse geometry')]
    params += [ChoiceParameter(visible_if='lc_datasets:<plural>', qualifier='lc_combine', value=kwargs.get('lc_combine', 'median'), choices=['median', 'max'], advanced=True, description='How to normalize each light curve prior to combining.')]

    params += [BoolParameter(qualifier='phase_bin', value=kwargs.get('phase_bin', True), description='Bin the input observations (see phase_nbins) if more than 2*phase_nbins.  NOTE: input observational sigmas will be ignored during binning.')]
    params += [IntParameter(qualifier='phase_nbins', visible_if='phase_bin:True', value=kwargs.get('phase_nbins', 500), limits=(100,None), description='Number of bins to use during phase binning input observations (will only be applied if len(times) > 2*phase_nbins)')]

    params += [ChoiceParameter(qualifier='orbit', value=kwargs.get('orbit', ''), choices=[''], description='Orbit to use for phasing the light curve referenced in the lc_datasets parameter')]

    params += [BoolParameter(qualifier='t0_near_times', value=kwargs.get('t0_near_times', True), description='Whether the returned value for t0_supconj should be forced to be in the range of the referenced observations.')]

    params += [BoolParameter(qualifier='expose_model', value=kwargs.get('expose_model', True), description='Whether to expose the 2-gaussian analytical models in the solution')]

    return ParameterSet(params)

def rv_geometry(**kwargs):
    """
    Create a <phoebe.parameters.ParameterSet> for solver options for the
    radial velocity geometry esimator.

    The input radial velocity datasets (`rv_datasets`) are combined without
    normalization.  These combined data are then used to estimate the
    semi-amplitude and `t0_supconj` which are then used to fit a Keplerian
    orbit using least-squares.  If RVs are available for both components,
    this results in proposed values for `t0_supconj`,
    `q`, `asini`, `ecc`, and `per0` for the corresponding `orbit` and `vgamma`
    of the system .  If RVs are only available for one of the components, then
    `q` is excluded, and the proposed `asini` is for the stellar component instead
    of the binary orbit.
    If `expose_model` is True, the analytical Keplerian RVs are exposed in the
    solution and available for
    plotting with <phoebe.parameters.ParameterSet.plot>.


    Generally, this will be used as an input to the kind argument in
    <phoebe.frontend.bundle.Bundle.add_solver>.  If attaching through
    <phoebe.frontend.bundle.Bundle.add_solver>, all `**kwargs` will be
    passed on to set the values as described in the arguments below.  Alternatively,
    see <phoebe.parameters.ParameterSet.set_value> to set/change the values
    after creating the Parameters.

    For example:

    ```py
    b.add_solver('estimator.rv_geometry')
    b.run_solver(kind='rv_geometry')
    ```

    Arguments
    ----------
    * `rv_datasets` (string or list, optional, default='*'): Radial velocity
        dataset(s) to use to extract RV geometry.
    * `phase_bin` (bool, optional, default=True): Bin the input observations (
        see `phase_nbins`) if more than 2*phase_nbins.  NOTE: input observational
        sigmas will be ignored during binning.
    * `phase_nbins` (int, optional, default=500): Number of bins to use during
        phase binning input observations
        (will only be applied if len(times) > 2*`phase_nbins`).  Only applicable
        if `phase_bin` is True.
    * `orbit` (string, optional, default=top-level orbit): Orbit to use for
        estimating orbital parameters.
    * `expose_model` (bool, optional, default=True): Whether to expose the
        Keplerian analytical models in the solution.

    Returns
    --------
    * (<phoebe.parameters.ParameterSet>): ParameterSet of all newly created
        <phoebe.parameters.Parameter> objects.
    """
    params = _comments_params(**kwargs)

    params += [SelectParameter(qualifier='rv_datasets', value=kwargs.get('rv_datasets', '*'), choices=[], description='Radial velocity dataset(s) to use to extract RV geometry')]
    params += [ChoiceParameter(qualifier='orbit', value=kwargs.get('orbit', ''), choices=[''], description='Orbit to use for estimating orbital parameters')]

    params += [BoolParameter(qualifier='phase_bin', value=kwargs.get('phase_bin', True), description='Bin the input observations (see phase_nbins) if more than 2*phase_nbins.  NOTE: input observational sigmas will be ignored during binning.')]
    params += [IntParameter(qualifier='phase_nbins', visible_if='phase_bin:True', value=kwargs.get('phase_nbins', 500), limits=(100,None), description='Number of bins to use during phase binning input observations (will only be applied if len(times) > 2*phase_nbins)')]

    # params += [BoolParameter(qualifier='t0_near_times', value=kwargs.get('t0_near_times', True), description='Whether the returned value for t0_supconj should be forced to be in the range of the referenced observations.')]

    params += [BoolParameter(qualifier='expose_model', value=kwargs.get('expose_model', True), description='Whether to expose the Keplerian analytical models in the solution')]


    return ParameterSet(params)


def ebai(**kwargs):
    """
    Create a <phoebe.parameters.ParameterSet> for solver options for the
    ebai artificial neural network solver.

    When using this solver, consider citing:
    * https://ui.adsabs.harvard.edu/abs/2008ApJ...687..542P

    See also:
    * <phoebe.frontend.bundle.Bundle.references>

    The input light curve datasets (`lc_datasets`) are each normalized
    according to `lc_combine`, combined and
    fitted with a 2 gaussian model which is then itself
    normalized and used as input to `ebai`.  Any necessary phase-shift required
    to ensure the primary is at a phase of 0 is used to provide the proposed
    value for `t0_supconj`.  The normalized 2 gaussian model is then sent through
    the matrix transformation for a pre-trained `ebai` artificial neural network
    resulting in proposed values for `teffratio`, `requivsumfrac`, `esinw`,
    `ecosw`, and `incl` for the corresponding `orbit`.

    Note that the current network only supports detached systems and will return
    all nans and a logger warning if either eclipse from the 2 gaussian model has
    a width greater than 0.25 (in phase-space).

    Generally, this will be used as an input to the kind argument in
    <phoebe.frontend.bundle.Bundle.add_solver>.  If attaching through
    <phoebe.frontend.bundle.Bundle.add_solver>, all `**kwargs` will be
    passed on to set the values as described in the arguments below.  Alternatively,
    see <phoebe.parameters.ParameterSet.set_value> to set/change the values
    after creating the Parameters.

    For example:

    ```py
    b.add_solver('estimator.ebai')
    b.run_solver(kind='ebai')
    ```

    Arguments
    ----------
    * `lc_datasets` (string or list, optional, default='*'): Light curve
        dataset(s) to pass to ebai.
    * `lc_combine` (string, optional, default='median'): How to normalize each
        light curve prior to combining.
    * `phase_bin` (bool, optional, default=True): Bin the input observations (
        see `phase_nbins`) if more than 2*phase_nbins.  NOTE: input observational
        sigmas will be ignored during binning.
    * `phase_nbins` (int, optional, default=500): Number of bins to use during
        phase binning input observations
        (will only be applied if len(times) > 2*`phase_nbins`).  Only applicable
        if `phase_bin` is True.
    * `orbit` (string, optional, default=top-level orbit): Orbit to use for
        phasing the light curve referenced in the `lc_datasets` parameter

    Returns
    --------
    * (<phoebe.parameters.ParameterSet>): ParameterSet of all newly created
        <phoebe.parameters.Parameter> objects.
    """
    params = _comments_params(**kwargs)

    params += [SelectParameter(qualifier='lc_datasets', value=kwargs.get('lc_datasets', '*'), choices=[], description='Light curve dataset(s) to pass to ebai')]
    params += [ChoiceParameter(visible_if='lc_datasets:<plural>', qualifier='lc_combine', value=kwargs.get('lc_combine', 'median'), choices=['median', 'max'], advanced=True, description='How to normalize each light curve prior to combining.')]

    params += [BoolParameter(qualifier='phase_bin', value=kwargs.get('phase_bin', True), description='Bin the input observations (see phase_nbins) if more than 2*phase_nbins.  NOTE: input observational sigmas will be ignored during binning.')]
    params += [IntParameter(qualifier='phase_nbins', visible_if='phase_bin:True', value=kwargs.get('phase_nbins', 500), limits=(100,None), description='Number of bins to use during phase binning input observations (will only be applied if len(times) > 2*phase_nbins)')]

    params += [ChoiceParameter(qualifier='orbit', value=kwargs.get('orbit', ''), choices=[''], description='Orbit to use for phasing the light curve referenced in the lc_datasets parameter')]

    return ParameterSet(params)





import re

from phoebe.parameters.twighelpers import _uniqueid_to_uniquetwig, _twig_to_uniqueid



class ConstraintVar(object):

    def __init__(self, bundle, twig):

        self._bundle = bundle
        self._parameter = None

        self._is_param = False
        self._is_constant = False
        self._is_method = False

        self._user_label = twig

        try:
            #~ print "ConstraintVar.__init__ twig: {}".format(twig)
            unique_label = _twig_to_uniqueid(bundle, twig=twig)
        except ValueError:
            # didn't return a unique match - either several or None

            # TODO: check length of results?
            self._is_param = False
            self._unique_label = None


            # TODO: check to see if valid constant, method, etc
            raise ValueError("provided twig does not result in unique match")

        else:
            # we have a unique match to a parameter
            self._is_param = True
            self._unique_label = unique_label


        if self._is_param:
            self._safe_label = ''.join(char for char in self._unique_label if char.isalpha())
            # update_user_label will also call _set_curly_label
            self.update_user_label()

    def _set_curly_label(self):
        """
        sets curly label based on user label
        """
        self._curly_label = '{'+self.user_label+'}'

    def update_user_label(self):
        """
        finds this parameter and gets the least_unique_twig from the bundle

        """
        self._user_label = _uniqueid_to_uniquetwig(self._bundle, self.unique_label)
        self._set_curly_label()

    @property
    def is_param(self):
        """
        is this variable a parameter of the system (vs a constant or method of an object)
        """
        return self._is_param

    #~ @property
    #~ def is_constant(self):
        #~ """
        #~ is this variable a constant (vs a parameter of the system or method of an object)
        #~ """
        #~ return self._is_constant
        #~
    #~ @property
    #~ def is_method(self):
        #~ """
        #~ is this variable a method of an object (vs a constant or a parameter of the system)
        #~ """
        #~ return self._is_method

    def get_parameter(self):
        """
        get the parameter object from the system for this var

        needs to be backend safe (not passing or storing bundle)
        """
        if not self.is_param:
            raise ValueError("this var does not point to a parameter")

        # this is quite expensive, so let's cache the parameter object so we only
        # have to filter on the first time this is called
        if self._parameter is None:
            self._parameter = self._bundle.get_parameter(uniqueid=self.unique_label, check_visible=False, check_default=False)

        return self._parameter

    def get_quantity(self, units=None, t=None):
        """
        """
        if self.is_param:
            # TODO: CAREFUL, this may cause infinite loops if we try to run constraints through get_value
            try:
                return self._bundle.get_quantity(uniqueid=self.unique_label, units=units, t=t, check_visible=False, check_default=False)
            except AttributeError:
                # then not a FloatParameter
                return self._bundle.get_value(uniqueid=self.unique_label, check_visible=False, check_default=False)

        else:
            # TODO: constants and methods
            raise NotImplementedError

    def get_value(self, units=None, t=None):
        """
        get the value (either of the constant or from the parameter) for this var

        needs to be backend safe (not passing or storing bundle)
        """
        if self.is_param:
            # TODO: CAREFUL, this may cause infinite loops if we try to run constraints through get_value
            return self._bundle.get_value(uniqueid=self.unique_label, units=units, t=t, check_visible=False, check_default=False)

        else:
            # TODO: constants and methods
            raise NotImplementedError


    @property
    def unique_label(self):
        """
        unique_label corresponding to parameter.get_unique_label
        call get_parameter(system) to retrieve the parameter itself


        needs to be backend safe (not passing or storing bundle)
        """
        return self._unique_label

    @property
    def user_label(self):
        """
        label as the user provided it

        needs to be backend safe (not passing or storing bundle)
        """
        return self._user_label

    @property
    def curly_label(self):
        """
        label with {} brackets - used for the user's view of the equation

        needs to be backend safe (not passing or storing bundle)
        """
        return self._curly_label

    @property
    def safe_label(self):
        """
        label that is safe to be passed through sympy - with escapes from any mathetmatical symbols

        needs to be backend safe (not passing or storing bundle)
        """
        return self._safe_label






def _uniqueid_to_uniquetwig(bundle, uniqueid):
    """
    """
    # print "*** parameters.twighelpers._uniqueid_to_uniquetwig uniqueid: {}".format(uniqueid)
    return bundle.get_parameter(uniqueid=uniqueid, check_visible=False, check_default=False).uniquetwig

def _twig_to_uniqueid(bundle, twig, **kwargs):
    """
    kwargs are passed on to filter
    """
    res = bundle.filter(twig=twig, force_ps=True, check_visible=False, check_default=False, **kwargs)
    # we force_ps instead of checking is_instance(res, ParameterSet) to avoid
    # having to import from parameters
    if len(res) == 1:
        #~ print "twighelpers._twig_to_uniqueid len(res): {}, res: {}".format(len(res), res)
        return res.to_list()[0].uniqueid
    else:
        raise ValueError("did not return a single unique match to a parameter for '{}'".format(twig))



import numpy as np
from phoebe.distortions import roche as _roche
from phoebe.backend import mesh as _mesh
from phoebe.dependencies import distl as _distl
import libphoebe
from scipy.optimize import newton, bisect

import logging
logger = logging.getLogger("BUILTIN")
logger.addHandler(logging.NullHandler())

# expose these at top-level so they're available to constraints
from numpy import sin, cos, tan, arcsin, arccos, arctan, arctan2, sqrt, log10

def phases_to_times(phase, period_choice, period, period_anom, dpdt_choice, dpdt, t0_choice, t0_supconj, t0_perpass, t0_ref):
    # print("*** builtin.phases_to_times period_choice: {}, t0_choice: {}".format(period_choice, t0_choice))
    if period_choice == 'period':
        period = period
    elif period_choice == 'period_anom':
        period = period_anom
    else:
        raise NotImplementedError()

    if t0_choice == 't0_supconj':
        t0 = t0_supconj
    elif t0_choice == 't0_perpass':
        t0 = t0_perpass
    elif t0_choice == 't0_ref':
        t0 = t0_ref
    else:
        raise NotImplementedError()

    if dpdt_choice == 'dpdt':
        dpdt = dpdt
    elif dpdt_choice == 'none':
        dpdt = 0.0
    else:
        raise NotImplementedError()

    # print("*** builtin.phases_to_times period: {}, t0: {}".format(period, t0))

    if isinstance(phase, list):
        phase = np.asarray(phase)

    if dpdt != 0:
        # NOTE: this seems to be incorrect and giving ridiculous answers
        time = t0 + 1./dpdt*(np.exp(dpdt*(phase))-period)
    else:
        time = t0 + (phase)*period

    return time

def times_to_phases(time, period_choice, period, period_anom, dpdt_choice, dpdt, t0_choice, t0_supconj, t0_perpass, t0_ref):
    # print("*** builtin.times_to_phases", period_choice, t0_choice)
    if period_choice == 'period':
        period = period
    elif period_choice == 'period_anom':
        period = period_anom
    else:
        raise NotImplementedError()

    if t0_choice == 't0_supconj':
        t0 = t0_supconj
    elif t0_choice == 't0_perpass':
        t0 = t0_perpass
    elif t0_choice == 't0_ref':
        t0 = t0_ref
    else:
        raise NotImplementedError()

    if dpdt_choice == 'dpdt':
        dpdt = dpdt
    elif dpdt_choice == 'none':
        dpdt = 0.0
    else:
        raise NotImplementedError()

    if isinstance(time, list):
        time = np.asarray(time)


    if dpdt != 0:
        phase = np.mod(1./dpdt * np.log(period + dpdt*(time-t0)), 1.0)
    else:
        phase = np.mod((time-t0)/period, 1.0)

    if isinstance(phase, float):
        if phase > 0.5:
            phase -= 1
    else:
        # then should be an array
        phase[phase > 0.5] -= 1

    return phase


def requiv_L1(q, syncpar, ecc, sma, incl_star, long_an_star, incl_orb, long_an_orb, compno, **kwargs):
    """
    """
    d = 1-ecc # compute at periastron
    true_anom = 0.0 # compute at periastron

    spin_xyz = _mesh.spin_in_system(incl_star, long_an_star)
    s = _mesh.spin_in_roche(spin_xyz, true_anom, long_an_orb, incl_orb)

    logger.debug("roche.roche_misaligned_critical_requiv(q={}, F={}, d={}, s={}, scale={})".format(q, syncpar, d, s, sma))
    q = _roche.q_for_component(q, compno)
    critical_requiv = _roche.roche_misaligned_critical_requiv(q, syncpar, d, s, sma)

    return critical_requiv

def potential_contact_L1(q, **kwargs):
    """
    """
    crit_pots = libphoebe.roche_critical_potential(q, 1., 1.)
    return crit_pots['L1']

def potential_contact_L23(q, **kwargs):
    """
    """
    crit_pots = libphoebe.roche_critical_potential(q, 1., 1.)
    return max(crit_pots['L2'], crit_pots['L3'])

def requiv_contact_L1(q, sma, compno, **kwargs):
    """
    for the contact case we can make the assumption of aligned, synchronous, and circular
    """
    return requiv_L1(q=q, syncpar=1, ecc=0, sma=sma, incl_star=0, long_an_star=0, incl_orb=0, long_an_orb=0, compno=compno, **kwargs)

def requiv_contact_L23(q, sma, compno, **kwargs):
    """
    for the contact case we can make the assumption of aligned, synchronous, and circular
    """
    logger.debug("requiv_contact_L23(q={}, sma={}, compno={})".format(q, sma, compno))
    crit_pot_L23 = potential_contact_L23(q)

    logger.debug("libphoebe.roche_contact_neck_min(phi=pi/2, q={}, d=1., crit_pot_L23={})".format(q, crit_pot_L23))
    nekmin = libphoebe.roche_contact_neck_min(np.pi/2., q, 1., crit_pot_L23)['xmin']
    # we now have the critical potential and nekmin as if we were the primary star, so now we'll use compno=0 regardless
    logger.debug("libphoebe.roche_contact_partial_area_volume(nekmin={}, q={}, d=1, Omega={}, compno={})".format(nekmin, q, crit_pot_L23, compno-1))
    crit_vol_L23 = libphoebe.roche_contact_partial_area_volume(nekmin, q, 1., crit_pot_L23, compno-1, lvolume=True, ldvolume=False, larea=False)['lvolume']

    logger.debug("resulting vol: {}, requiv: {}".format(crit_vol_L23, (3./4*1./np.pi*crit_vol_L23)**(1./3) * sma))

    return (3./4*1./np.pi*crit_vol_L23)**(1./3) * sma

def pot_to_fillout_factor(q, pot, **kwargs):
    # calling libphoebe.roche_critical_potential is fairly cheap, so we'll
    # just parameterize this directly with q and make multiple calls to
    # critical potentials rather than having a nightmare of logics
    # between constraints.

    pot_L1 = potential_contact_L1(q)
    pot_L23 = potential_contact_L23(q)
    return (pot - pot_L1) / (pot_L23 - pot_L1)

def fillout_factor_to_pot(q, fillout_factor, **kwargs):
    pot_L1 = potential_contact_L1(q)
    pot_L23 = potential_contact_L23(q)
    return fillout_factor * (pot_L23 - pot_L1) + pot_L1


def esinw2per0(ecc, esinw):
    """
    """
    # print "*** constraints.builtin.esinw2per0", ecc, esinw

    if ecc==0.:
        return 0.
    else:
        per0 = np.arcsin(esinw/ecc)
        if np.isnan(per0):
            raise ValueError("esinw={} and ecc={} results in nan for per0".format(esinw, ecc))
        return per0

def ecosw2per0(ecc, ecosw):
    """
    """
    # print "*** constraints.builtin.ecosw2per0", ecc, ecosw

    if ecc==0.:
        return 0.
    else:
        per0 = np.arccos(ecosw/ecc)
        if np.isnan(per0):
            raise ValueError("ecosw={} and ecc={} results in nan for per0".format(ecosw, ecc))
        return per0

def esinw2ecc(esinw, per0):
    """
    """
    # print("*** constraints.builtin.esinw2ecc", esinw, per0)
    if np.sin(per0) == 0:
        raise ValueError("esinw={} and per0={} results in nan for ecc".format(esinw, per0))
    ecc = esinw/np.sin(per0)
    if np.isnan(ecc):
        raise ValueError("esinw={} and per0={} results in nan for ecc".format(esinw, per0))
    return ecc

def ecosw2ecc(ecosw, per0):
    """
    """
    # print("*** constraints.builtin.ecosw2ecc", ecosw, per0)
    if np.cos(per0) == 0:
        raise ValueError("ecosw={} and per0={} results in nan for ecc".format(ecosw, per0))
    ecc = ecosw/np.cos(per0)
    if np.isnan(ecc):
        raise ValueError("ecosw={} and per0={} results in nan for ecc".format(ecosw, per0))
    return ecc



def _delta_t_supconj_perpass(t, period, ecc, per0, dpdt, dperdt, t0):
    """
    time shift between superior conjuction and periastron passage
    """
    period = period + (t-t0)*dpdt
    per0 = (per0 + (t-t0)*dperdt) % (2*np.pi)
    ups_sc = np.pi/2-per0
    E_sc = 2*np.arctan( np.sqrt((1-ecc)/(1+ecc)) * np.tan(ups_sc/2) )
    M_sc = E_sc - ecc*np.sin(E_sc)
    return period*(M_sc/2./np.pi)

def t0_perpass_to_supconj(t0_perpass, period, ecc, per0, dpdt, dperdt, t0):
    """
    """
    return t0_perpass + _delta_t_supconj_perpass(t0_perpass, period, ecc, per0, dpdt, dperdt, t0)

def t0_supconj_to_perpass(t0_supconj, period, ecc, per0, dpdt, dperdt, t0):
    """
    """
    return t0_supconj - _delta_t_supconj_perpass(t0_supconj, period, ecc, per0, dpdt, dperdt, t0)

def _delta_t_supconj_ref(t, period, ecc, per0, dpdt, dperdt, t0):
    """
    time shift between superior conjunction and reference time (time at
    which the primary star crosses the barycenter along line-of-sight)
    """
    period = period + (t-t0)*dpdt
    per0 = (per0 + (t-t0)*dperdt) % (2*np.pi)
    ups_sc = np.pi/2-per0
    E_sc = 2*np.arctan( np.sqrt((1-ecc)/(1+ecc)) * np.tan(ups_sc/2) )
    M_sc = E_sc - ecc*np.sin(E_sc)
    return period*((M_sc+per0)/2./np.pi - 1./4)

def t0_ref_to_supconj(t0_ref, period, ecc, per0, dpdt, dperdt, t0):
    """
    """
    return t0_ref + _delta_t_supconj_ref(t0_ref, period, ecc, per0, dpdt, dperdt, t0)

def t0_supconj_to_ref(t0_supconj, period, ecc, per0, dpdt, dperdt, t0):
    """
    """
    return t0_supconj - _delta_t_supconj_ref(t0_supconj, period, ecc, per0, dpdt, dperdt, t0)

def requiv_to_pot_contact(requiv, q, sma, compno=1):
    """
    :param requiv: user-provided equivalent radius
    :param q: mass ratio
    :param sma: semi-major axis (d = sma because we explicitly assume circular orbits for contacts)
    :param compno: 1 for primary, 2 for secondary
    :return: potential and fillout factor
    """
    logger.debug("requiv_to_pot_contact(requiv={}, q={}, sma={}, compno={})".format(requiv, q, sma, compno))

    # since the functions called here work with normalized r, we need to set d=D=sma=1.
    # or provide sma as a function parameter and normalize r here as requiv = requiv/sma
    requiv = requiv/sma
    vequiv = 4./3*np.pi*requiv**3
    d = 1.
    F = 1.

    logger.debug("libphoebe.roche_contact_Omega_at_partial_vol(vol={}, phi=pi/2, q={}, d={}, choice={})".format(vequiv, q, d, compno-1))
    return libphoebe.roche_contact_Omega_at_partial_vol(vequiv, np.pi/2, q, d, choice=compno-1)

def pot_to_requiv_contact(pot, q, sma, compno=1):
    """
    """
    logger.debug("pot_to_requiv_contact(pot={}, q={}, sma={}, compno={})".format(pot, q, sma, compno))
    d = 1.
    F = 1.
    crit_pots = libphoebe.roche_critical_potential(q, d, F)
    crit_pot_L1 = crit_pots['L1']
    crit_pot_L23 = max(crit_pots['L2'], crit_pots['L3'])
    if pot > crit_pot_L1:
        raise ValueError("potential > L1 critical value")
    if pot < crit_pot_L23:
        raise ValueError("potential < L2/L3 critical value")

    try:
        logger.debug("libphoebe.roche_contact_neck_min(pi/2, q={}, d={}, pot={})".format(q, d, pot))
        nekmin = libphoebe.roche_contact_neck_min(np.pi / 2., q, d, pot)['xmin']
        logger.debug("libphoebe.roche_contact_partial_area_volume(nekmin={}, q={}, d={}, pot={}, compno={})".format(nekmin, q, d, pot, compno-1))
        volume_equiv = libphoebe.roche_contact_partial_area_volume(nekmin, q, d, pot, compno-1, lvolume=True, ldvolume=False, larea=False)['lvolume']
        # returns normalized vequiv, should multiply by sma back for requiv in SI
        return sma * (3./4 * 1./np.pi * volume_equiv)**(1./3)
    except:
        # replace this with actual check in the beginning or before function call
        raise ValueError('potential probably out of bounds for contact envelope')

def distl_from_json(json):
    return _distl.from_json(json)



import numpy as np
# from numpy import cos, sin, tan, pi, sqrt
from math import sqrt, sin, cos, acos, atan2, trunc, pi

# from phoebe import c
import libphoebe

# from scipy.optimize import newton

import logging
logger = logging.getLogger("ROCHE")
logger.addHandler(logging.NullHandler())

def q_for_component(q, component=1):
    """
    Flip the mass-ratio `q` if secondary component.

    Arguments
    -------------
    * `q` (float): the mass-ratio of the system/orbit (defined as m2/m1)
    * `component` (int, optional, default=1): whether the requested component
        is the primary (1) or secondary (2) component.

    Returns
    -----------
    * (float) `q` if `component` is 1, or `1/q` if component is 2

    Raises
    -----------
    * NotImplementedError: if `component` is neither 1 or 2.
    """
    if component==1:
        return q
    elif component==2:
        return 1./q
    else:
        raise NotImplementedError

def pot_for_component(pot, q, component=1, reverse=False):
    """
    Handle converting the potential into the star's frame of reference.

    Note: `q` should already be in the frame of the requested star.  If necessary,
    pass the output from <phoebe.distortions.roche.q_for_component>.

    If `component` is 1:
        * `pot` is returned as is

    If `component` is 2:
        * if `reverse` is True: `pot/q + 0.5 * (q-1)/q`
        * if `reverse` is False: `q*pot - 0.5 * (q-1)`

    Arguments
    ------------
    * `pot` (float): the equipotential of the star as defined as if it were the
        primary component.
    * `q` (float): the mass-ratio in the frame of `component`.  If necessary,
        call <phoebe.distortions.roche.q_for_component> first.
    * `component` (int, optional, default=1): whether the requested component
        is the primary (1) or secondary (2) component.
    * `reverse` (bool, optional, default=False): whether to do the reverse operation.
        If True, then `pot` is assumed to be in the reference frame of `component`
        and is returned as if in the reference from of the primary component.

    Returns
    ----------
    * (float): the converted equipotential.

    Raises
    ------------
    * NotImplementedError: if `component` is neither 1 or 2.)
    """
    # currently only used by legacy wrapper: consider moving/removing
    if component==1:
        return pot
    elif component==2:
        if reverse:
            return pot/q + 0.5*(q-1)/q
        else:
            return q*pot - 0.5 * (q-1)
    else:
        raise NotImplementedError

def roche_misaligned_critical_requiv(q, F, d, s, scale=1.0):
    """
    Compute the critical value for the equivalent radius for the misaligned
    roche case.

    Note: all inputs and output are floats, so it is important to pass values
    with consistent units.

    Note: `q` should already be in the frame of the requested star.  If necessary,
    pass the output from <phoebe.distortions.roche.q_for_component>.

    Note: `s` should be in roche coordinates and at the requested time at which
    the critical radius is being requested.  For the aligned case, `s` should
    be [0, 0, 1].

    Arguments
    ----------
    * `q` (float): the mass-ratio in the frame of `component`.  If necessary,
        call <phoebe.distortions.roche.q_for_component> first.
    * `F` (float): roche synchronicity parameter.
    * `d` (float): instantaneous separation between the two components in the
        orbit.
    * `s` (array of length 3): rotation spin vector, in roche coordinates, at
        the requested time.
    * `scale` (float, optional, default=1.0): value by which to scale the output
        (to convert from roche to real units, for example)

    Returns
    -----------
    * (float): the critical equivalent radius.

    """
    volume_critical = libphoebe.roche_misaligned_critical_volume(q, F, d, s)
    logger.debug("libphoebe.roche_misaligned_critical_volume(q={}, F={}, d={}, s={}) => {}".format(q, F, d, s, volume_critical))

    requiv_critical = scale * (volume_critical * 3./4 * 1./np.pi)**(1./3)
    # logger.debug("roche.roche_misaligned_critical_requiv = {}".format(requiv_critical))

    return requiv_critical

def requiv_to_pot(requiv, sma, q, F, d, s=np.array([0.,0.,1.]), component=1):
    """
    Convert from equivalent radius to equipotential for the misaligned roche case.

    Note: all inputs and output are floats, so it is important to pass values
    with consistent units.

    Note: `q` should already be in the frame of the requested star.  If necessary,
    pass the output from <phoebe.distortions.roche.q_for_component>.

    Note: if `component` is not 1, the equipotential will be converted into the
    primary frame by calling <phoebe.distortions.roche.pot_for_component> with
    `reverse=True`.  To avoid this behavior, pass `component=1`.

    See also:
    * <phoebe.distortions.roche.pot_to_requiv>

    Arguments
    -----------
    * `requiv` (float): the equivalent radius
    * `sma` (float): the semi-major axis of the orbit
    * `q` (float): the mass-ratio in the frame of `component`.  If necessary,
        call <phoebe.distortions.roche.q_for_component> first.
    * `F` (float): the synchronicity parameter.
    * `d` (float): instantaneous separation between the two components in the
        orbit.
    * `s` (array of length 3): rotation spin vector, in roche coordinates, at
        the requested time.
    * `component` (int, optional, default=1): whether the requested component
        is the primary (1) or secondary (2) component.

    Returns
    ----------
    * (float): the equipotential in the primary frame.
    """
    logger.debug("roche.requiv_to_pot(requiv={}, sma={}, q={}, F={}, d={}, s={}, component={})".format(requiv, sma, q, F, d, s, component))
    volume = 4./3 * np.pi * requiv**3 / sma**3
    logger.debug("roche_misaligned_Omega_at_vol(volume={}, q={}, F={}, d={}, s={})".format(volume, q, F, d, s))
    Phi = libphoebe.roche_misaligned_Omega_at_vol(volume,
                                                  q, F, d, s)

    return pot_for_component(Phi, q, component=component, reverse=True)

def pot_to_requiv(pot, sma, q, F, d, s=np.array([0.,0.,1.]), component=1):
    """
    Convert from equipotential to equivalent radius for the misaligned roche case.

    Note: all inputs and output are floats, so it is important to pass values
    with consistent units.

    Note: if `component` is not 1, the equipotential `pot` will be first converted into the
    star's frame by calling <phoebe.distortions.roche.pot_for_component> and
    the mass-ratio `q` will be converted to the star's frame via
    <phoebe.distortions.roche.q_for_component>.  To avoid this behavior, pass
    `component=1`.

    See also:
    * <phoebe.distortions.roche.requiv_to_pot>

    Arguments
    -----------
    * `pot` (float): the equipotential of the star as defined in the primary frame.
    * `sma` (float): the semi-major axis of the orbit.  Pass as `1` to
        return `requiv` in roche coordinates.
    * `q` (float): the mass-ratio in the frame of the primary star (m2/m1).
        Do not call <phoebe.distortions.roche.q_for_component> first as this
        is called internally if `component=2`.
    * `F` (float): the synchronicity parameter.
    * `d` (float): instantaneous separation between the two components in the
        orbit.
    * `s` (array of length 3): rotation spin vector, in roche coordinates, at
        the requested time.
    * `component` (int, optional, default=1): whether the requested component
        is the primary (1) or secondary (2) component.

    Returns
    ----------
    * (float): the equivalent radius.
    """
    logger.debug("roche.pot_to_requiv(pot={}, sma={}, q={}, F={}, d={}, s={}, component={})".format(pot, sma, q, F, d, s, component))
    q = q_for_component(q, component)

    pot = pot_for_component(pot, q, component)
    logger.debug("libphoebe.roche_area_volume(q={}, F={}, d={}, Omega={})".format(q, F, d, pot))
    volume = libphoebe.roche_misaligned_area_volume(q, F, d, s, pot,
                                                    choice=0,
                                                    lvolume=True,
                                                    larea=False)['lvolume']

    # convert from roche units to scaled (solar) units
    volume *= sma**3

    # now convert from volume (in solar units) to requiv
    requiv = (volume * 3./4 * 1./np.pi)**(1./3)

    return requiv

def rpole_to_pot_aligned(rpole, sma, q, F, d, component=1):
    """
    Transforms from polar radius to equipotential under the aligned roche case.

    Note: all inputs and output are floats, so it is important to pass values
    with consistent units.

    Note: if `component` is not 1, the mass-ratio `q` will be converted to the star's frame via
    <phoebe.distortions.roche.q_for_component> and the resulting equipotential
    will be converted back to the primary frame via
    <phoebe.distortions.roche.pot_for_component> with `reverse=True`.
    To avoid this behavior, pass `component=1`.

    See also:
    * <phoebe.distortions.roche.pot_to_rpole_aligned>

    Arguments
    -----------
    * `rpole` (float): polar radius
    * `sma` (float): semi-major axis of the orbit
    * `q` (float): the mass-ratio in the frame of the primary star (m2/m1).
        Do not call <phoebe.distortions.roche.q_for_component> first as this
        is called internally if `component=2`.
    * `F` (float): synchronicity parameter
    * `d` (float): instantaneous unitless distance between the two components
    * `component` (int, optional, default=1): whether the requested component
        is the primary (1) or secondary (2) component.

    Returns
    ---------
    * (float) the aligned roche equipotential
    """
    q = q_for_component(q, component=component)
    rpole_ = np.array([0, 0, rpole/sma])
    logger.debug("libphoebe.roche_Omega(q={}, F={}, d={}, rpole={})".format(q, F, d, rpole_))
    pot = libphoebe.roche_Omega(q, F, d, rpole_)
    return pot_for_component(pot, component, reverse=True)

def pot_to_rpole_aligned(pot, sma, q, F, d, component=1):
    """
    Transforms from equipotential to polar radius under the aligned roche case.

    Note: all inputs and output are floats, so it is important to pass values
    with consistent units.

    Note: if `component` is not 1, the mass-ratio `q` will be converted to the star's frame via
    <phoebe.distortions.roche.q_for_component> and the equipotential `pot`
    will be converted to the star's frame via
    <phoebe.distortions.roche.pot_for_component>.
    To avoid this behavior, pass `component=1`.

    See also:
    * <phoebe.distortions.roche.rpole_to_pot_aligned>

    Arguments
    -----------
    * `pot` (float): the aligned roche equipotential
    * `sma` (float): semi-major axis of the orbit.  To return polar radius
        in roche units, use `sma=1`.
    * `q` (float): the mass-ratio in the frame of the primary star (m2/m1).
        Do not call <phoebe.distortions.roche.q_for_component> first as this
        is called internally if `component=2`.
    * `F` (float): synchronicity parameter
    * `d` (float): instantaneous unitless distance between the two components
    * `component` (int, optional, default=1): whether the requested component
        is the primary (1) or secondary (2) component.

    Returns
    ---------
    * (float) the aligned roche equipotential
    """
    q = q_for_component(q, component=component)
    Phi = pot_for_component(pot, q, component=component)
    logger.debug("libphobe.roche_pole(q={}, F={}, d={}, Omega={})".format(q, F, d, pot))
    return libphoebe.roche_pole(q, F, d, pot) * sma

def requiv_to_rpole_aligned(requiv, sma, q, F, d, component=1):
    """
    Transforms from equivalent radius to polar radius under the aligned roche case.

    Note: all inputs and output are floats, so it is important to pass values
    with consistent units.

    This is just a helper function around <phoebe.distortions.roche.requiv_to_pot>
    and <phoebe.distortions.roche.pot_to_rpole_aligned>.

    See also:
    * <phoebe.distortions.roche.rpole_to_requiv_aligned>

    Arguments
    -----------
    * `requiv` (float): the equivalent radius
    * `sma` (float): semi-major axis of the orbit.
    * `q` (float): the mass-ratio in the frame of the primary star (m2/m1).
        Do not call <phoebe.distortions.roche.q_for_component> first as this
        is called internally if `component=2`.
    * `F` (float): synchronicity parameter
    * `d` (float): instantaneous unitless distance between the two components
    * `component` (int, optional, default=1): whether the requested component
        is the primary (1) or secondary (2) component.

    Returns
    ---------
    * (float) the polar radius
    """
    pot = requiv_to_pot(requiv, sma, q, F, d, component=component)
    return pot_to_rpole_aligned(pot, sma, q, F, d, component=component)

def rpole_to_requiv_aligned(rpole, sma, q, F, d, component=1):
    """
    Transforms from equivalent radius to polar radius under the aligned roche case.

    Note: all inputs and output are floats, so it is important to pass values
    with consistent units.

    This is just a helper function around <phoebe.distortions.roche.rpole_to_pot_aligned>
    and <phoebe.distortions.roche.pot_to_requiv>.

    See also:
    * <phoebe.distortions.roche.requiv_to_rpole_aligned>

    Arguments
    -----------
    * `rpole` (float): the polar radius
    * `sma` (float): semi-major axis of the orbit.
    * `q` (float): the mass-ratio in the frame of the primary star (m2/m1).
        Do not call <phoebe.distortions.roche.q_for_component> first as this
        is called internally if `component=2`.
    * `F` (float): synchronicity parameter
    * `d` (float): instantaneous unitless distance between the two components
    * `component` (int, optional, default=1): whether the requested component
        is the primary (1) or secondary (2) component.

    Returns
    ---------
    * (float) the equivalent radius
    """
    pot = rpole_to_pot_aligned(rpole, sma, q, F, d, component=component)
    return pot_to_requiv(pot, sma, q, F, d, component=component)



def BinaryRoche (r, D, q, F, Omega=0.0):
    r"""
    Computes a value of the asynchronous, eccentric Roche potential.

    If :envvar:`Omega` is passed, it computes the difference.

    The asynchronous, eccentric Roche potential is given by [Wilson1979]_

    .. math::

        \Omega = \frac{1}{\sqrt{x^2 + y^2 + z^2}} + q\left(\frac{1}{\sqrt{(x-D)^2+y^2+z^2}} - \frac{x}{D^2}\right) + \frac{1}{2}F^2(1+q)(x^2+y^2)

    @param r:      relative radius vector (3 components)
    @type r: 3-tuple
    @param D:      instantaneous separation
    @type D: float
    @param q:      mass ratio
    @type q: float
    @param F:      synchronicity parameter
    @type F: float
    @param Omega:  value of the potential
    @type Omega: float
    """
    return 1.0/sqrt(r[0]*r[0]+r[1]*r[1]+r[2]*r[2]) + q*(1.0/sqrt((r[0]-D)*(r[0]-D)+r[1]*r[1]+r[2]*r[2])-r[0]/D/D) + 0.5*F*F*(1+q)*(r[0]*r[0]+r[1]*r[1]) - Omega

def dBinaryRochedx (r, D, q, F):
    """
    Computes a derivative of the potential with respect to x.

    @param r:      relative radius vector (3 components)
    @param D:      instantaneous separation
    @param q:      mass ratio
    @param F:      synchronicity parameter
    """
    return -r[0]*(r[0]*r[0]+r[1]*r[1]+r[2]*r[2])**-1.5 -q*(r[0]-D)*((r[0]-D)*(r[0]-D)+r[1]*r[1]+r[2]*r[2])**-1.5 -q/D/D + F*F*(1+q)*r[0]

def d2BinaryRochedx2(r, D, q, F):
    """
    Computes second derivative of the potential with respect to x.

    @param r:      relative radius vector (3 components)
    @param D:      instantaneous separation
    @param q:      mass ratio
    @param F:      synchronicity parameter
    """
    return (2*r[0]*r[0]-r[1]*r[1]-r[2]*r[2])/(r[0]*r[0]+r[1]*r[1]+r[2]*r[2])**2.5 +\
          q*(2*(r[0]-D)*(r[0]-D)-r[1]*r[1]-r[2]*r[2])/((r[0]-D)*(r[0]-D)+r[1]*r[1]+r[2]*r[2])**2.5 +\
          F*F*(1+q)

def dBinaryRochedy (r, D, q, F):
    """
    Computes a derivative of the potential with respect to y.

    @param r:      relative radius vector (3 components)
    @param D:      instantaneous separation
    @param q:      mass ratio
    @param F:      synchronicity parameter
    """
    return -r[1]*(r[0]*r[0]+r[1]*r[1]+r[2]*r[2])**-1.5 -q*r[1]*((r[0]-D)*(r[0]-D)+r[1]*r[1]+r[2]*r[2])**-1.5 + F*F*(1+q)*r[1]

def  dBinaryRochedz(r, D, q, F):
    """
    Computes a derivative of the potential with respect to z.

    @param r:      relative radius vector (3 components)
    @param D:      instantaneous separation
    @param q:      mass ratio
    @param F:      synchronicity parameter
    """
    return -r[2]*(r[0]*r[0]+r[1]*r[1]+r[2]*r[2])**-1.5 -q*r[2]*((r[0]-D)*(r[0]-D)+r[1]*r[1]+r[2]*r[2])**-1.5

def dBinaryRochedr (r, D, q, F):
    """
    Computes a derivative of the potential with respect to r.

    @param r:      relative radius vector (3 components)
    @param D:      instantaneous separation
    @param q:      mass ratio
    @param F:      synchronicity parameter
    """

    r2 = (r*r).sum()
    r1 = np.sqrt(r2)

    return -1./r2 - q*(r1-r[0]/r1*D)/((r[0]-D)*(r[0]-D)+r[1]*r[1]+r[2]*r[2])**1.5 - q*r[0]/r1/D/D + F*F*(1+q)*(1-r[2]*r[2]/r2)*r1

def Lag1(q):
    # L1
    dxL = 1.0
    L1 = 1e-3
    while abs(dxL) > 1e-6:
        dxL = - dBinaryRochedx([L1, 0.0, 0.0], 1., q, 1.) / d2BinaryRochedx2([L1, 0.0, 0.0], 1., q, 1.)
        L1 = L1 + dxL

    return L1



from phoebe.parameters import *
from phoebe.parameters.unit_choices import unit_choices as _unit_choices
import numpy as np

from .common import _use_autofig, _mplcmaps, _mplcolors, _mplmarkers, _mpllinestyles, MPLPropCycler, _label_units_lims, _figure_style_sources, _figure_style_nosources, _figure_uncover_highlight_animate

def lc_periodogram(b, **kwargs):
    params = []

    params += [ChoiceParameter(qualifier='solution', value=kwargs.get('solution', ''), choices=[''], description='Solution to include in the plot')]

    kwargs.setdefault('color', 'black' if _use_autofig else None)
    kwargs.setdefault('marker', 'None' if _use_autofig else None)
    kwargs.setdefault('linestyle', 'solid' if _use_autofig else None)
    params += _figure_style_nosources(b, **kwargs)

    # params += [BoolParameter(qualifier='legend', value=kwargs.get('legend', True), advanced=True, description='Whether to draw the legend')]

    return ParameterSet(params)

def rv_periodogram(b, **kwargs):
    params = []

    params += [ChoiceParameter(qualifier='solution', value=kwargs.get('solution', ''), choices=[''], description='Solution to include in the plot')]

    kwargs.setdefault('color', 'black' if _use_autofig else None)
    kwargs.setdefault('marker', 'None' if _use_autofig else None)
    kwargs.setdefault('linestyle', 'solid' if _use_autofig else None)
    params += _figure_style_nosources(b, **kwargs)

    # params += [BoolParameter(qualifier='legend', value=kwargs.get('legend', True), advanced=True, description='Whether to draw the legend')]

    return ParameterSet(params)

def lc_geometry(b, **kwargs):
    params = []

    params += [ChoiceParameter(qualifier='solution', value=kwargs.get('solution', ''), choices=[''], description='Solution to include in the plot')]

    # kwargs.setdefault('color', 'black' if _use_autofig else None)
    # kwargs.setdefault('marker', 'None' if _use_autofig else None)
    # kwargs.setdefault('linestyle', 'solid' if _use_autofig else None)
    # params += _figure_style_nosources(b, **kwargs)

    # params += [BoolParameter(qualifier='legend', value=kwargs.get('legend', True), advanced=True, description='Whether to draw the legend')]

    return ParameterSet(params)

def rv_geometry(b, **kwargs):
    params = []

    params += [ChoiceParameter(qualifier='solution', value=kwargs.get('solution', ''), choices=[''], description='Solution to include in the plot')]

    # kwargs.setdefault('color', 'black' if _use_autofig else None)
    # kwargs.setdefault('marker', 'None' if _use_autofig else None)
    # kwargs.setdefault('linestyle', 'solid' if _use_autofig else None)
    # params += _figure_style_nosources(b, **kwargs)

    # params += [BoolParameter(qualifier='legend', value=kwargs.get('legend', True), advanced=True, description='Whether to draw the legend')]

    return ParameterSet(params)

def ebai(b, **kwargs):
    params = []

    params += [ChoiceParameter(qualifier='solution', value=kwargs.get('solution', ''), choices=[''], description='Solution to include in the plot')]

    # kwargs.setdefault('color', 'black' if _use_autofig else None)
    # kwargs.setdefault('marker', 'None' if _use_autofig else None)
    # kwargs.setdefault('linestyle', 'solid' if _use_autofig else None)
    # params += _figure_style_nosources(b, **kwargs)

    # params += [BoolParameter(qualifier='legend', value=kwargs.get('legend', True), advanced=True, description='Whether to draw the legend')]

    return ParameterSet(params)


def emcee(b, **kwargs):
    params = []

    params += [ChoiceParameter(qualifier='solution', value=kwargs.get('solution', ''), choices=[''], description='Solution to include in the plot')]

    params += [ChoiceParameter(qualifier='style', value=kwargs.get('style', 'lnprobability'), choices=['corner', 'failed', 'lnprobability', 'walks'], description='style of plot')]

    # TODO: implement y for walks (need to set choices based on the solution fitted_twigs)
    # params += [ChoiceParameter(visible_if='style:walks', qualifier='y', value=kwargs.get('y', ''), choices=[''], description='Parameter samples to plot along y-axis')]

    # params += _label_units_lims('y', visible_if='style:lnprobability', default_unit=u.dimensionless_unscaled, is_default=True, **kwargs)

    kwargs.setdefault('color', 'black' if _use_autofig else None)
    kwargs.setdefault('marker', 'None' if _use_autofig else None)
    kwargs.setdefault('linestyle', 'solid' if _use_autofig else None)
    params += _figure_style_nosources(b, **kwargs)

    # params += [BoolParameter(qualifier='legend', value=kwargs.get('legend', True), advanced=True, description='Whether to draw the legend')]

    return ParameterSet(params)

def dynesty(b, **kwargs):
    params = []

    params += [ChoiceParameter(qualifier='solution', value=kwargs.get('solution', ''), choices=[''], description='Solution to include in the plot')]

    params += [ChoiceParameter(qualifier='style', value=kwargs.get('style', 'run'), choices=['corner', 'run', 'trace'], description='style of plot')]

    # params += _label_units_lims('y', visible_if='style:lnprobability', default_unit=u.dimensionless_unscaled, is_default=True, **kwargs)

    # kwargs.setdefault('color', 'black' if _use_autofig else None)
    # kwargs.setdefault('marker', 'None' if _use_autofig else None)
    # kwargs.setdefault('linestyle', 'solid' if _use_autofig else None)
    # params += _figure_style_nosources(b, **kwargs)

    # params += [BoolParameter(qualifier='legend', value=kwargs.get('legend', True), advanced=True, description='Whether to draw the legend')]

    return ParameterSet(params)

# del deepcopy
# del _add_component, _add_dataset, _label_units_lims, _run_compute
# del download_passband, list_installed_passbands, list_online_passbands, list_passbands, parameter_from_json, parse_json, send_if_client, update_if_client
# del fnmatch



"""
Environment variables while manually building/installing:

PHOEBE_SKIP_SCRIPTS=TRUE: will not install phoebe-server and phoebe-autofig executables (removing the following dependencies: flask, flask-cors, flask-socketio, gevent-websocket)
"""
import sys

try:
  import numpy
except ImportError:
  print("Numpy is needed for running and building of PHOEBE")
  sys.exit(1)

from numpy.distutils.core import setup, Extension
from numpy.distutils.command.build_ext import build_ext
from numpy.distutils.command.build_py import build_py

from distutils.version import LooseVersion, StrictVersion
from distutils.cmd import Command

import platform
import os
import re


# Python version checks (in both __init__.py and setup.py)
if sys.version_info[0] == 3:
    if sys.version_info[1] < 6:
        raise ImportError("PHOEBE supports python 3.6+")
elif sys.version_info[0] == 2:
    if sys.version_info[1] < 7:
        raise ImportError("PHOEBE 2.3+ requires python 3.6+")
else:
    raise ImportError("PHOEBE supports python 3.6+")

#
# Auxiliary functions
#

def removefile(f):
  try:
    os.remove(f)
  except OSError:
    pass


def find_version_gcc(s):
  if len(s):
    for i in s.split():
      if re.search("^\d+\.\d+\.*\d*$",i): return i
  return ''


def __find_version_clang(s):
  ver = ''
  sp = s.split()

  for i, w in enumerate(sp):
    if w == 'version':
      if i < len(sp): ver = sp[i+1]
      break
  return ver

def find_version_clang(s):
  if re.search(r'LLVM version', s):
    ver = ("llvm", __find_version_clang(s))
  else:
    ver = ("clang", __find_version_clang(s))
  return ver

def find_version_intel(s):
  return s.split()[-2]

#
# Check the platform and C++ compiler (g++ > 5.0)
#
def check_unix_compiler(plat, plat_ver, compiler, extensions, compiler_name):

  status = False

  s = os.popen(compiler_name + " --version").readline().strip()

  # debug output
  print("Compiler: %s"%(compiler_name))
  print("Compiler version: %s"%(s))

  compiler_found = False;
  version_ok = False;

  # GCC compiler
  if re.search(r'gcc', compiler_name) or re.search(r'^g\+\+', compiler_name):
    name = 'gcc'
    compiler_found = True
    ver = find_version_gcc(s)
    if ver != '': version_ok = LooseVersion(ver) >= LooseVersion("5.0")

  # LLVm clang compiler
  elif re.search(r'^clang', compiler_name):
    name = 'clang'
    compiler_found = True

    # https://stackoverflow.com/questions/19774778/when-is-it-necessary-to-use-use-the-flag-stdlib-libstdc
    if plat == 'Darwin':
      opt ="-stdlib=libc++"
      if LooseVersion(plat_ver) < LooseVersion("13.0"): #OS X Mavericks
        for e in extensions:
          if not (opt in e.extra_compile_args):
            e.extra_compile_args.append(opt)

    ver = find_version_clang(s)

    if ver != '':
      if ver[0] == 'clang': # CLANG version
        version_ok = LooseVersion(ver[1]) >= LooseVersion("3.3")
      else:                 # LLVM version
        version_ok = LooseVersion(ver[1]) >= LooseVersion("7.0")

  # Intel compilers
  elif re.search(r'^icc', compiler_name) or re.search(r'^icpc', compiler_name):
    name = 'icc'
    compiler_found = True

    ver = find_version_intel(s)
    version_ok = LooseVersion(ver) >= LooseVersion("16")

  # compiler could be masquerading under different name
  # check this out:
  #  ln -s `which gcc` a
  #  CC=`pwd`/a python check_compiler.py

  if not compiler_found:

    import tempfile
    import random
    import string

    tempdir = tempfile.gettempdir();

    pat = ''.join(random.choice(string.ascii_uppercase + string.digits) for _ in range(10))
    src = pat+'_compiler_check.c'
    exe = pat+'_compiler_check.exe'
    obj = pat+'_compiler_check.o'

    with open(tempdir + '/' + src, 'w') as tmp:
      tmp.writelines(
        ['#include <stdio.h>\n',
         'int main(int argc, char *argv[]) {\n',
          '#if defined (__INTEL_COMPILER)\n',
          '  printf("icc %d.%d", __INTEL_COMPILER, __INTEL_COMPILER_UPDATE);\n',
          '#elif defined(__clang__)\n',
          '  printf("clang %d.%d.%d", __clang_major__, __clang_minor__, __clang_patchlevel__);\n',
          '#elif defined(__GNUC__)\n',
          '  printf("gcc %d.%d.%d\\n",__GNUC__,__GNUC_MINOR__,__GNUC_PATCHLEVEL__);\n',
          '#else\n',
          '  printf("not_gcc");\n',
          '#endif\n',
          'return 0;\n',
          '}\n'
        ])

    try:
      objects = compiler.compile([tempdir+'/'+ src], output_dir='/')
      compiler.link_executable(objects, exe, output_dir = tempdir)

      out = os.popen(tempdir+'/'+ exe).read()

      if len(out) != 0:
        name, ver = out.split(' ')

        if name == 'gcc':
          version_ok = LooseVersion(ver) >= LooseVersion("5.0")
          compiler_found = True

        if name == 'clang':
          version_ok = LooseVersion(ver) >= LooseVersion("3.3") # not LLVM version !!!
          compiler_found = True

        if name == 'icc':
          version_ok = LooseVersion(ver) >= LooseVersion("1600")
          compiler_found = True
    except:
      print("Unable to build a test program to determine the compiler.")
      status = False

    # Cleanup
    removefile(tempdir+'/'+ src)
    removefile(tempdir+'/'+ exe)
    removefile(tempdir+'/'+ obj)

  if compiler_found:
    if version_ok:
      print("Ready to compile with %s %s." % (name, ver))
      status = True
    else:
      print("Compiler is too old. PHOEBE requires gcc 5.0, clang 3.3, or icc 1600 or above.\nThe found compiler is %s %s." % (name, ver))
      status = False
  else:
    print("Did not recognize the compiler %s." % (compiler_name))
    status = False

  return status

#
# Hooking the building of extentions
#
class build_check(build_ext):

  def build_extensions(self):

    plat = platform.system()
    plat_ver = platform.release();

    print("OS: %s" %(plat))
    print("OS version: %s" %(plat_ver))

    if plat == 'Windows':

      print("On windows we don't perform checks")
      build_ext.build_extensions(self)

    elif plat in ['Linux', 'Darwin']:
      if (
          check_unix_compiler(plat, plat_ver, self.compiler, self.extensions, self.compiler.compiler_cxx[0]) and
          check_unix_compiler(plat, plat_ver, self.compiler, self.extensions, self.compiler.compiler_so[0])
         ):

        for e in self.extensions:
          print("  extra_args=%s"%(e.extra_compile_args))

        build_ext.build_extensions(self)
      else:
        print("Cannot build phoebe2. Please check the dependencies and try again.")
        sys.exit(1)
    else:
      print("Unknown architecture, so no pre-checks done. Please report in the case of build failure.")
      sys.exit(1)
#
# Setting up the external modules
#

class import_check(Command):
  description = "Checks python modules needed to successfully import phoebe"
  user_options = []

  def initialize_options(self):
    pass

  def finalize_options(self):
    pass

  def run(self):
    required, optional = [], []
    try:
      import astropy
      astropy_version = astropy.__version__
      if LooseVersion(astropy_version) < LooseVersion('1.0'):
        required.append('astropy 1.0+')
    except:
      required.append('astropy')
    try:
      import scipy
      scipy_version = scipy.__version__
      if LooseVersion(scipy_version) < LooseVersion('0.1'):
        required.append('scipy 0.1+')
    except:
      required.append('scipy')
    try:
      import matplotlib
      mpl_version = matplotlib.__version__
      if LooseVersion(mpl_version) < LooseVersion('1.4.3'):
        optional.append('matplotlib 1.4.3+')
    except:
      optional.append('matplotlib')
    try:
      import corner
      corner_version = corner.__version__
      if LooseVersion(corner_version) < LooseVersion('2.0.0'):
        optional.append('corner 2.0+')
    except:
      optional.append('corner')
    try:
      import sympy
      sympy_version = sympy.__version__
      if LooseVersion(sympy_version) < LooseVersion('1.0'):
        optional.append('sympy 1.0+')
    except:
      optional.append('sympy')

    if required == []:
      print('All required import dependencies satisfied.')
    else:
      print('NOTE: while all the build dependencies are satisfied, the following import dependencies')
      print('      are still missing: %s.' % required)
      print('      You will not be able to import phoebe before you install those dependencies.')

    if optional == []:
      print('All optional import dependencies satisfied.')
    else:
      print('NOTE: while all the build dependencies are satisfied, the following optional dependencies')
      print('      are still missing: %s.' % optional)
      print('      Some of the core phoebe functionality will be missing until you install those dependencies.')

class PhoebeBuildCommand(build_py):
  def run(self):
    build_py.run(self)
    self.run_command('build_ext')
    self.run_command('check_imports')

ext_modules = [
    Extension('libphoebe',
      sources = ['phoebe/lib/libphoebe.cpp'],
      language='c++',
      extra_compile_args = ["-std=c++11"],
      include_dirs=[numpy.get_include()]
      ),

    Extension('phoebe.algorithms.ceclipse',
      language='c++',
      sources = ['phoebe/algorithms/ceclipse.cpp'],
      include_dirs=[numpy.get_include()]
      ),
]

#
# Main setup
#
def _env_variable_bool(key, default):
    value = os.getenv(key, default)
    if isinstance(value, bool):
        return value
    elif value.upper()=='TRUE':
        return True
    else:
        return False

# numpy distutils does not support passing long_description_content_type
# so we'll have to manually convert to rst and strip out the html in the header
# but we also don't want to have to make m2r a hard dependency for installation
try:
    import m2r
except ImportError:
    with open('README.md', 'r') as fh:
        long_description = fh.read()
else:
    long_description_s = m2r.parse_from_file("README.md").split('\n')
    long_description = "\n".join(long_description_s[long_description_s.index("INTRODUCTION"):])

setup (name = 'phoebe',
       version = '2.3.16',
       description = 'PHOEBE 2.3.16',
       long_description=long_description,
       author = 'PHOEBE development team',
       author_email = 'phoebe-devel@lists.sourceforge.net',
       url = 'http://github.com/phoebe-project/phoebe2',
       classifiers=[
            'Development Status :: 5 - Production/Stable',
            'Intended Audience :: Science/Research',
            'Topic :: Software Development :: Build Tools',
            "Natural Language :: English",
            "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
            "Topic :: Scientific/Engineering",
            "Topic :: Software Development :: Libraries :: Python Modules",
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8',
            'Programming Language :: Python :: 3 :: Only',
        ],
       python_requires='>=3.6, <4',
       download_url = 'https://github.com/phoebe-project/phoebe2/tarball/2.3.16',
       packages = ['phoebe', 'phoebe.parameters', 'phoebe.parameters.solver', 'phoebe.parameters.figure', 'phoebe.frontend', 'phoebe.constraints', 'phoebe.dynamics', 'phoebe.distortions', 'phoebe.algorithms', 'phoebe.atmospheres', 'phoebe.backend', 'phoebe.solverbackends', 'phoebe.solverbackends.ebai', 'phoebe.utils', 'phoebe.helpers', 'phoebe.pool', 'phoebe.dependencies', 'phoebe.dependencies.autofig', 'phoebe.dependencies.nparray', 'phoebe.dependencies.distl', 'phoebe.dependencies.unitsiau2015'],
       install_requires=['numpy>=1.12','scipy>=1.2','astropy>=1.0', 'corner', 'pytest', 'requests', 'python-socketio[client]']+['flask', 'flask-cors', 'flask-socketio', 'gevent-websocket'],
       package_data={'phoebe.atmospheres':['tables/wd/*', 'tables/passbands/*'],
                     'phoebe.frontend':['default_bundles/*.bundle'],
                     'phoebe.solverbackends.ebai': ['*.data', '*.weights']
                    },
       ext_modules = ext_modules,
       scripts=None if  _env_variable_bool('PHOEBE_SKIP_SCRIPTS', False) else ['client-server/phoebe-server', 'client-server/phoebe-autofig'],
       cmdclass = {
         'build_ext': build_check,
         'check_imports': import_check,
         'build_py': PhoebeBuildCommand
        }
      )
